/**@file

@brief This library contains functions called from other unCPC scripts.

@author Jonas Arroyo (BE-ICS-FD)
@author Ivan Koblik (IT-CO-FE)
@author Nikolay Kulman (IT-CO-FE)
@author Geraldine Thomas (IT-CO-FE)
@author Francisco Mico (IT-CO-FE)
@author Alexey Merezhin (EN-ICE-PLC)

@copyright
        &copy;Copyright CERN 2013 - all rights reserved

@par Creation Date
  14.11.2003

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI, CTRL
*/


const string cpcConfigGenericFunctions_sPARAM_DPE_VALUE = "value:"; // used to specify the dpe's value instead of addressing


/** Check data of the $sDeviceType object before configuration. It calls or reproduces following calls:
        unConfigGenericFunctions_checkParameters(dsConfigs, sDeviceType, dsExceptionInfo);
        unConfigGenericFunctions_checkDeviceConfig(dsConfigs, sDeviceType, dsExceptionInfo);

Device can specify its own checkConfig function by specifying it in device type configuration.
Also it can overwrite the check for specific front-end as in following example for CMW front end of CPC_WordStatus object:
{code}
CMW_CPC_WordStatus_checkConfig(dyn_string configLine, dyn_string &dsExceptionInfo) {
    cpcConfigGenericFunctions_checkDeviceConfig(UN_CONFIG_CPC_WORDSTATUS_DPT_NAME, configLine, dsExceptionInfo);
}
{code}

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param dsExceptionInfo exception info

*/

void cpcConfigGenericFunctions_genericCheckDeviceConfig(dyn_string dsConfigs, dyn_string &dsExceptionInfo)
{
  dyn_string dsDollar, dsExceptionInfoTemp;
  string sFrontEndType, sFunction;


  // Initialize CPC constants (performance issue)
  cpcConfigGenericFunctions_initConstants();

  if( dynlen(dsConfigs) >= UN_CONFIG_COMMON_LENGTH )
  {
    // 1. Common

    unConfigGenericFunctions_checkParameters(dsConfigs, $sDeviceType, dsExceptionInfo);

    // check if there's special implementation or go with default one
    if( unGenericObject_shapeExists("TextFieldPlcType") )
    {
      sFrontEndType = TextFieldPlcType.text;
    }
    else
    {
      sFrontEndType = $sFrontEndType;
      dsDollar      = makeDynString("$sFrontEndType:"       + $sFrontEndType,
                                    "$dsArchiveList:"       + $dsArchiveList,
                                    "$sFrontEndVersion:"    + $sFrontEndVersion,
                                    "$bCheckBeforeImport:"  + $bCheckBeforeImport,
                                    "$sFrontEndDeviceType:" + $sFrontEndDeviceType);
    }

    sFrontEndType = unImportDevice_getFrontEndDeviceType(sFrontEndType);

    sFunction = sFrontEndType + "_" + $sDeviceType + UN_CONFIG_DEVICE_CHECKFUNCTIONSUFFIX;

    if( isFunctionDefined(sFunction) )
    {
      evalScript(dsExceptionInfoTemp, "dyn_string main(dyn_string dsConfigs) "                                                      +
                                      "{"                                                                                           +
                                      "  dyn_string dsExceptionInfo;"                                                               +
                                      "  if( isFunctionDefined(\"" + sFunction + "\"))"                                             +
                                      "  {"                                                                                         +
                                      "    " + sFunction + "(dsConfigs, dsExceptionInfo);"                                          +
                                      "  }"                                                                                         +
                                      "  else"                                                                                      +
                                      "  {"                                                                                         +
                                      "    fwException_raise(dsExceptionInfo,"                                                      +
                                      "                      \"ERROR\","                                                            +
                                      "                      \"" + sFunction + "\" + getCatStr(\"unGeneration\",\"UNKNOWNCHECK\")," +
                                      "                      \"\");"                                                                +
                                      "  }"                                                                                         +
                                      "  return dsExceptionInfo;"                                                                   +
                                      "}",
                                      dsDollar,
                                      dsConfigs);
    }
    else
    {
      cpcConfigGenericFunctions_checkDeviceConfig($sDeviceType, dsConfigs, dsExceptionInfo);
    }

  dynAppend(dsExceptionInfo, dsExceptionInfoTemp);

  }
  else
  {
    fwException_raise(dsExceptionInfo, "ERROR", "CPC_WordStatusConfig_checkConfig: " + getCatStr("unGeneration", "BADINPUTS"), "");
    return;
  }

}




/** Creates dp (or checks if it is already presented) and sets configuration for the $sDeviceType object. That calls or reproduces following calls:
        unConfigGenericFunctions_createDp
        unConfigGenericFunctions_setCommon
        unConfigGenericFunctions_setDeviceConfig

Device can specify its own setConfig function by specifying it in device type configuration.
Also it can overwrite the set for specific front-end as in following example for Schneider (_UnPlc) front end of CPC_WordStatus object:
{code}
_UnPlc_CPC_WordStatus_setConfig(dyn_string dsConfigs, dyn_string &exceptionInfo) {
    cpcConfigGenericFunctions_setDeviceConfig(dsConfigs, exceptionInfo);
}
{code}

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param dsExceptionInfo exception info
*/

void cpcConfigGenericFunctions_genericSetDeviceConfig(dyn_string dsConfigs, dyn_string &dsExceptionInfo)
{
  dyn_string dsDollar, dsExceptionInfoTemp;
  string sDpName, sFrontEndType, sFunction;


  // Initialize CPC constants (performance issue)
  cpcConfigGenericFunctions_initConstants();

  // 1. Initialization : checking + dpCreate
  unConfigGenericFunctions_checkAll(dsConfigs, $sDeviceType, dsExceptionInfo);
  if( dynlen(dsExceptionInfo) <= 0 )
  {
    // Good inputs
    sDpName = dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME];
    unConfigGenericFunctions_createDp(sDpName, $sDeviceType, dsExceptionInfo);
    if( dynlen(dsExceptionInfo) <= 0 )
    {
      // Creation ok or dp already exists
      // 2. Generate device

      // 2.1. Parameters
      unConfigGenericFunctions_setCommon(dsConfigs, dsExceptionInfo);

      // 2.2 call the device check function
      if( unGenericObject_shapeExists("TextFieldPlcType") )
      {
        sFrontEndType = TextFieldPlcType.text;
      }
      else
      {
        sFrontEndType = $sFrontEndType;
        dsDollar      = makeDynString("$sFrontEndType:"       + $sFrontEndType,
                                      "$dsArchiveList:"       + $dsArchiveList,
                                      "$sFrontEndVersion:"    + $sFrontEndVersion,
                                      "$bCheckBeforeImport:"  + $bCheckBeforeImport,
                                      "$sFrontEndDeviceType:" + $sFrontEndDeviceType);
      }

      if( isDollarDefined("$bIgnoreDriverRelatedConfigs") )
      {
        dynAppend(dsDollar, "$bIgnoreDriverRelatedConfigs:" + $bIgnoreDriverRelatedConfigs);
      }

      sFrontEndType = unImportDevice_getFrontEndDeviceType(sFrontEndType);
      unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "unImportDevice", "unConfigGenericFunctions_setDeviceConfig", sFrontEndType, $sDeviceType, dsDollar, "-");

      sFunction = sFrontEndType + "_" + $sDeviceType + UN_CONFIG_DEVICE_SETFUNCTIONSUFFIX;
      // DebugTN("cpcConfigGenericFunctions_genericSetDeviceConfig() -> sFunction = " + sFunction + ", isDefined = " + isFunctionDefined(sFunction));

      if( isFunctionDefined(sFunction) )
      {
        evalScript(dsExceptionInfoTemp, "dyn_string main(dyn_string configLine)"                                                     +
                                        "{"                                                                                          +
                                        "  dyn_string dsExceptionInfo;"                                                              +
                                        "  if( isFunctionDefined(\"" + sFunction + "\") )"                                           +
                                        "  {"                                                                                        +
                                        "    " + sFunction + "(configLine, dsExceptionInfo);"                                        +
                                        "  }"                                                                                        +
                                        "  else"                                                                                     +
                                        "  {"                                                                                        +
                                        "    fwException_raise(dsExceptionInfo,"                                                     +
                                        "                      \"ERROR\","                                                           +
                                        "                      " + sFunction + "\" \" + getCatStr(\"unGeneration\",\"UNKNOWNSET\")," +
                                        "                      \"\");"                                                               +
                                        "  }"                                                                                        +
                                        "  return dsExceptionInfo;"                                                                  +
                                        "}",
                                        dsDollar,
                                        dsConfigs);
      }
      else
      {
        cpcConfigGenericFunctions_setDeviceConfig(dsConfigs, dsExceptionInfo);
      }

      dynAppend(dsExceptionInfo, dsExceptionInfoTemp);

    }
  }

}




/**Generic function to initialize CPC constants for import process

Initialization function to cache CPC constants during the import process

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param None
@return None

*/


void cpcConfigGenericFunctions_initConstants()
{
  dyn_string dsExceptionInfo;


  if( globalExists("g_mCpcConst") == FALSE )
  {
    // DebugTN("cpcConfigGenericFunctions_initConstants() -> creating global g_mCpcConst");
    if( cpcConfigGenericFunctions_createGlobalConst(dsExceptionInfo) == FALSE )
    {
      DebugTN("cpcConfigGenericFunctions_getConstantByName() -> Error, creating global variable: " + g_mCpcConst + ", due to: " + dsExceptionInfo[2]);
      return;
    }

    // Initialize device constants
    CPC_AnaDigConfig_initializeConstants();
    CPC_AnaDOConfig_initializeConstants();
    CPC_AnalogAlarmConfig_initializeConstants();
    CPC_AnalogConfig_initializeConstants();
    CPC_AnalogInputConfig_initializeConstants();
    CPC_AnalogOutputConfig_initializeConstants();
    CPC_AnalogParameterConfig_initializeConstants();
    CPC_AnalogStatusConfig_initializeConstants();
    CPC_ControllerConfig_initializeConstants();
    CPC_DigitalAlarmConfig_initializeConstants();
    CPC_DigitalInputConfig_initializeConstants();
    CPC_DigitalOutputConfig_initializeConstants();
    CPC_DigitalParameterConfig_initializeConstants();
    CPC_LocalConfig_initializeConstants();
    CPC_MassFlowControllerConfig_initializeConstants();
    CPC_OnOffConfig_initializeConstants();
    CPC_ProcessControlObjectConfig();
    CPC_SteppingMotorConfig_initializeConstants();
    CPC_Word2AnalogStatusConfig_initializeConstants();
    CPC_WordParameterConfig_initializeConstants();
    CPC_WordStatusConfig_initializeConstants();
    cpcRcpBuffersConfig_initializeConstants();
  }

}




/**Generic function that check device config.

Check validness of importation line using getConfig(), getParamNames() functions and importation constants defined in the device config script.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type
@param dsConfigs importation line, processed in the unCore
@param dsExceptionInfo exception info

*/
void cpcConfigGenericFunctions_checkDeviceConfig(string sDeviceType, dyn_string dsConfigs, dyn_string &dsExceptionInfo)
{
  int iDeviceLength, iAddressPosition, iNormalPos, iAcknowledgeAlarmPosition, iDefaultDeadbandPos, iDefaultDeadbandTypePos, iDefaultRangeMaxPos, iDefaultRangeMinPos, iDefaultUnitPos, iDefaultFormatPos, iDescPos, iArchivePos, iTimeFilterPos, iLLLimitPos, iLLimitPos, iHLimitPos, iHHLimitPos, iParametersPos, iAddressIdx, iParametersIdx, iSmsCategoriesPos, iMaskEventPos, iAlarmActivePos, iAddressLen, iParametersLen;
  string sFrontEndType, sCurrentAddressName, sAddressValue, sTimeFilter, sDpe, sCustomFunction;
  bool bHasArchive, bAcknowledgeAlarmValue, bMaskEventValue;
  mapping mDeviceAddresses, mCurrentAddressProps;
  dyn_string dsDeviceAddressNames, dsParameters, dsParameterPair, dsSmsCategories;


  if( dynlen(dsConfigs) < UN_CONFIG_COMMON_LENGTH )
  {
    // totally broken config
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkDeviceConfig: " + sDeviceType + " importation line is less that minimum", "");
  }

  // check length
  if( !cpcConfigGenericFunctions_checkConfigLength(sDeviceType, dsConfigs, bHasArchive) )
  {
    iDeviceLength = UN_CONFIG_COMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPLength(sDeviceType) + cpcConfigGenericFunctions_getConstantDPAdditionalLength(sDeviceType);
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkDeviceConfig: " + sDeviceType + " expected length " + (iDeviceLength + 2) + "(+3) differs from config length " + (dynlen(dsConfigs) + 2), "");
    return;
  }

  // check archives
  if( bHasArchive )
  {
    cpcConfigGenericFunctions_checkArchiveConfig(sDeviceType, dsConfigs, dsExceptionInfo);
    if( dynlen(dsExceptionInfo) > 0 )
    {
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkDeviceConfig: " + sDeviceType + " " + getCatStr("unGeneration", "BADINPUTS"), "");
      return;
    }
  }

  sFrontEndType = _cpcConfigGenericFunctions_getFrontEndType();

  // check device specific params
  mDeviceAddresses = cpcConfigGenericFunctions_getDPEConfig(sDeviceType);
  dsDeviceAddressNames = mappingKeys(mDeviceAddresses);

  // check addresses
  iAddressLen = dynlen(dsDeviceAddressNames);
  for( iAddressIdx = 1 ; iAddressIdx <= iAddressLen ; iAddressIdx++ )
  {
    sCurrentAddressName  = dsDeviceAddressNames[iAddressIdx];
    mCurrentAddressProps = mDeviceAddresses[sCurrentAddressName];
    if( !cpcConfigGenericFunctions_isOldSoftFe(sFrontEndType) )
    {
      if( mCurrentAddressProps["dataType"] == CPC_BOOL )
      {
        if( mCurrentAddressProps["bitPosition"] < 0 ||
            mCurrentAddressProps["bitPosition"] > 15  )
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sCurrentAddressName + " has wrong bit position", "");
        }
      }
      else
      {
        iAddressPosition = cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sCurrentAddressName);
        if( iAddressPosition < 1 )
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkDeviceAddress: " + sDeviceType + " has no position constant for " + sCurrentAddressName, "");
          return;
        }

        sAddressValue = dsConfigs[UN_CONFIG_COMMON_LENGTH + iAddressPosition];
        cpcConfigGenericFunctions_checkDeviceAddress(dsConfigs, sFrontEndType, sDeviceType, sAddressValue, mCurrentAddressProps, bHasArchive, dsExceptionInfo);
      }

      if( dynlen(dsExceptionInfo) > 0 )
      {
        return;
      }
    }

    if( mCurrentAddressProps["isAlarm"] )
    {
      // check alarm
      iNormalPos = cpcConfigGenericFunctions_getConstant(sDeviceType, "NORMAL_POSITION");
      if( iNormalPos > 0 )
      {
        unConfigGenericFunctions_checkNormalPosition(dsConfigs[UN_CONFIG_COMMON_LENGTH + iNormalPos], dsExceptionInfo);
      }

      if( dynlen(dsExceptionInfo) > 0 )
      {
        return;
      }

      if( mCurrentAddressProps["hasAcknowledge"] )
      {
        iAcknowledgeAlarmPosition = cpcConfigGenericFunctions_getConstant(sDeviceType, "ACK_ALARM");
        if( iAcknowledgeAlarmPosition > 0 )
        {
          cpcConfigGenericFunctions_checkBool("ACK_ALARM", dsConfigs[UN_CONFIG_COMMON_LENGTH + iAcknowledgeAlarmPosition], bAcknowledgeAlarmValue, dsExceptionInfo);
        }
        else
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no acknowledge type for " + sCurrentAddressName, "");
        }
      }

      if( dynlen(dsExceptionInfo) > 0 )
      {
        return;
      }
    }


    if( mCurrentAddressProps["hasSmooth"] )
    {
      iDefaultDeadbandPos     = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "DEADBAND");
      iDefaultDeadbandTypePos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "DEADBAND_TYPE");
      if( iDefaultDeadbandPos > 0 && iDefaultDeadbandTypePos > 0 )
      {
        unConfigGenericFunctions_checkDeadBand(dsConfigs[UN_CONFIG_COMMON_LENGTH + iDefaultDeadbandPos], dsConfigs[UN_CONFIG_COMMON_LENGTH + iDefaultDeadbandTypePos], dsExceptionInfo);
      }
      else
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no deadband for " + sCurrentAddressName, "");
      }
    }


    if( mCurrentAddressProps["hasPvRange"] )
    {
      iDefaultRangeMaxPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "RANGEMAX");
      iDefaultRangeMinPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "RANGEMIN");
      if( iDefaultRangeMaxPos > 0 && iDefaultRangeMinPos > 0 )
      {
        unConfigGenericFunctions_checkRange(dsConfigs[UN_CONFIG_COMMON_LENGTH + iDefaultRangeMinPos], dsConfigs[UN_CONFIG_COMMON_LENGTH + iDefaultRangeMaxPos], dsExceptionInfo);
      }
      else
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no range for " + sCurrentAddressName, "");
      }
    }


    if( mCurrentAddressProps["hasUnit"] )
    {
      iDefaultUnitPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "UNIT");
      if( iDefaultUnitPos > 0 )
      {
        // @TODO: not checked ever
      }
      else
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no unit for " + sCurrentAddressName, "");
      }
    }


    if( mCurrentAddressProps["hasFormat"] )
    {
      iDefaultFormatPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "FORMAT");
      if( iDefaultFormatPos > 0 )
      {
        unConfigGenericFunctions_checkFormat(dsConfigs[UN_CONFIG_COMMON_LENGTH + iDefaultFormatPos], dsExceptionInfo);
      }
      else
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no format for " + sCurrentAddressName, "");
      }
    }


    if( mCurrentAddressProps["hasMapping"] )
    {
      iDescPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "MAPPING");
      if( iDescPos > 0 )
      {
        cpcConfigGenericFunctions_checkMapping(dsConfigs, dsConfigs[UN_CONFIG_COMMON_LENGTH + iDescPos], dsExceptionInfo);
      }
      else
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no description for " + sCurrentAddressName, "");
      }
    }


    if( mCurrentAddressProps["hasArchive"] )
    {
      if( (strpos(sCurrentAddressName, "evStsReg") == -1) && (mCurrentAddressProps["dataType"] != CPC_BOOL) )
      {
        iArchivePos    = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "ARCHIVE_ACTIVE");
        iTimeFilterPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sCurrentAddressName, bHasArchive, "ARCHIVE_TIME_FILTER");
        if( iArchivePos > 0 )
        {
          sTimeFilter = "";
          if( iTimeFilterPos > 0 )
          {
            sTimeFilter = dsConfigs[UN_CONFIG_COMMON_LENGTH + iTimeFilterPos];

            // @TODO: current checks are not full and covers only special case of time filter
            unConfigGenericFunctions_checkArchiveParameters(sTimeFilter, dsConfigs[UN_CONFIG_COMMON_LENGTH + iArchivePos], dsExceptionInfo);
          }
        }
        else
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no archive for " + sCurrentAddressName, "");
        }
      }
    }


    if( mCurrentAddressProps["5RangesAlerts"] )
    {
      iAcknowledgeAlarmPosition = cpcConfigGenericFunctions_getConstant(sDeviceType, "ACK_ALARM");
      if( iAcknowledgeAlarmPosition > 0 )
      {
        cpcConfigGenericFunctions_checkBool("ACK_ALARM", dsConfigs[UN_CONFIG_COMMON_LENGTH + iAcknowledgeAlarmPosition], bAcknowledgeAlarmValue, dsExceptionInfo);
        if( dynlen(dsExceptionInfo) > 0 )
        {
          return;
        }
      }
      else
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no acknowledge type for " + sCurrentAddressName, "");
      }

      iLLLimitPos     = cpcConfigGenericFunctions_getConstant(sDeviceType, "LLLIMIT");
      iLLimitPos      = cpcConfigGenericFunctions_getConstant(sDeviceType, "LLIMIT");
      iHLimitPos      = cpcConfigGenericFunctions_getConstant(sDeviceType, "HLIMIT");
      iHHLimitPos     = cpcConfigGenericFunctions_getConstant(sDeviceType, "HHLIMIT");
      iAlarmActivePos = cpcConfigGenericFunctions_getConstant(sDeviceType, "ALARM_ACTIVE");
      if( iAlarmActivePos > 0 &&
          iLLLimitPos     > 0 &&
          iLLimitPos      > 0 &&
          iHLimitPos      > 0 &&
          iHHLimitPos     > 0   )
      {
        cpcConfigGenericFunctions_check5RangesAlert(dsConfigs,
                                                    iAlarmActivePos,
                                                    makeDynInt(iLLLimitPos,
                                                               iLLimitPos,
                                                               iHLimitPos,
                                                               iHHLimitPos),
                                                    dsExceptionInfo);
      }
      else
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkImportConfig: " + sDeviceType + " has no alarm active or not all limits for " + sCurrentAddressName, "");
      }
    }

    if( dynlen(dsExceptionInfo) > 0 )
    {
      return;
    }
  }


  iParametersPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "PARAMETERS");
  if( iParametersPos > 0 )
  {
    cpcConfigGenericFunctions_checkParameters(sDeviceType, dsConfigs[UN_CONFIG_COMMON_LENGTH + iParametersPos], dsExceptionInfo);
    if( dynlen(dsExceptionInfo) > 0 )
    {
      return;
    }

    dsParameters = strsplit(dsConfigs[UN_CONFIG_COMMON_LENGTH + iParametersPos], ",");
    iParametersLen = dynlen(dsParameters);
    for( iParametersIdx = 1 ; iParametersIdx <= iParametersLen ; iParametersIdx++ )
    {
      dsParameterPair = strsplit(dsParameters[iParametersIdx], "=");
      if( strpos(dsParameterPair[1], "_addr") > 0 )
      {
        sDpe = dsParameterPair[1];
        strreplace(sDpe, "_addr", "");
        cpcConfigGenericFunctions_checkDeviceAddress(dsConfigs, sFrontEndType, sDeviceType, dsParameterPair[2], mDeviceAddresses[sDpe], bHasArchive, dsExceptionInfo);
      }
    }
  }


  iSmsCategoriesPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "SMS_CATEGORIES");
  if( iSmsCategoriesPos > 0 )
  {
    dsSmsCategories = strsplit(dsConfigs[UN_CONFIG_COMMON_LENGTH + iSmsCategoriesPos], ",");
    unProcessAlarm_checkCategory(dsSmsCategories, dsExceptionInfo);
  }
  if( dynlen(dsExceptionInfo) > 0 )
  {
    return;
  }


  iMaskEventPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "MASKEVENT");
  if( iMaskEventPos > 0 )
  {
    cpcConfigGenericFunctions_checkBool("MASKEVENT", dsConfigs[UN_CONFIG_COMMON_LENGTH + iMaskEventPos], bMaskEventValue, dsExceptionInfo);
  }
  if( dynlen(dsExceptionInfo) > 0 )
  {
    return;
  }


  sCustomFunction = "CPC_";
  if( cpcConfigGenericFunctions_isOldSoftFe(sFrontEndType) )
  {
    sCustomFunction += "SOFT_FE_";
  }
  sCustomFunction += cpcConfigGenericFunctions_getShortTypeName(sDeviceType) + "Config_checkCustomConfig";
  if( isFunctionDefined(sCustomFunction) )
  {
    evalScript(dsExceptionInfo, "dyn_string main(dyn_string configLine, bool bHasArchive)"          +
                              "{"                                                                 +
                              "  dyn_string dsExceptionInfo;"                                       +
                              "  " + sCustomFunction + "(configLine, bHasArchive, dsExceptionInfo);" +
                              "  return dsExceptionInfo;"                                           +
                              "}",
                              makeDynString("$sFrontEndDeviceType:" + $sFrontEndDeviceType),
                              dsConfigs,
                              bHasArchive);
    if( dynlen(dsExceptionInfo) > 0 )
    {
      return;
    }
  }

}




/**Check if string represents a boolean value.

@param sName name of the field, used in error message
@param sSource string to check
@param bRes boolean representation of the string
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_checkBool(string sName, string sSource, bool &bRes, dyn_string &dsExceptionInfo)
{
  sSource = strtolower(strrtrim(strltrim(sSource)));

  if( sSource == "true" ||
      sSource == "1"    ||
      sSource == "y"    ||
      sSource == "yes"    )
  {
    bRes = true;
  }
  else
  {
    if( sSource == "false" ||
        sSource == "0"     ||
        sSource == "n"     ||
        sSource == "no"      )
    {
      bRes = false;
    }
    else
    {
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkBool: the value '" + sSource + "' is not correct for the " + sName, "");
    }
  }
}




/**Check comma separated string for the correctness.

Verifies that each of the values is a pair of an identifier, assignment sign, and an integer as a value.
For example: "value1=23,value2=10"

@param dsConfigs importation line, processed in the unCore
@param sLabels string to test that contains comma separated values
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_checkMapping(dyn_string dsConfigs, string sLabels, dyn_string &dsExceptionInfo)
{
  dyn_string dsLabels, dsValue, dsExceptionInfoTmp;
  int iLabelIdx, iValue, iLabelLen, iCorrectValue;


  if( sLabels == "-" )
  {
    sLabels = ""; // backward compatibility after UCPC-1341
  }

  // if the string is not empty, check that each element is: anInt=aString
  dsLabels = strsplit(sLabels, ",");

  // Check that each of the sub string is "n=string"
  iLabelLen = dynlen(dsLabels);
  for( iLabelIdx = 1 ; iLabelIdx <= iLabelLen ; iLabelIdx++ )
  {
    dsValue = strsplit(dsLabels[iLabelIdx], "=");
    if( dynlen(dsValue) != 2 )
    {
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkMapping:" + getCatStr("unGeneration", "BADDATA"), "");
      return;
    }
    else
    {
      //dsValue first element should be the integer
      unConfigGenericFunctions_checkInt(strrtrim(strltrim(dsValue[1])), iCorrectValue, dsExceptionInfoTmp);
      if( dynlen(dsExceptionInfoTmp) > 0 )
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkMapping:" + getCatStr("unGeneration", "BADDATA"), "");
        return;
      }
    }
  }
}




/**Check if the configs lenght is correct. Also reports if config has archiving in it.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType DPT name
@param dsConfigs importation line, processed in the unCore
@param bHasArchive set to true if passed config importation line contains bool/analog/event archives, otherwise set to false

@return result of checking, true if config has a proper length, otherwise it's false.

*/

bool cpcConfigGenericFunctions_checkConfigLength(string sDeviceType, dyn_string dsConfigs, bool &bHasArchive)
{
  int iDeviceLength, iConfigLength;


  iDeviceLength = UN_CONFIG_COMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPLength(sDeviceType) + cpcConfigGenericFunctions_getConstantDPAdditionalLength(sDeviceType);
  iConfigLength = dynlen(dsConfigs);

  if( _cpcConfigGenericFunctions_getFrontEndType() == "SOFT_FE"               &&
      (iConfigLength != iDeviceLength                                      &&
       iConfigLength != (iDeviceLength + UN_CONFIG_ADDITIONAL_ARCHIVE_LENGTH))  )
  {
    if( !mappingHasKey(g_mCpcConst, "NEW_SOFT_FE") ||
        !g_mCpcConst["NEW_SOFT_FE"]                   )
    {
      // Setting new SOFT_FE version
      g_mCpcConst["NEW_SOFT_FE"] = true;
    }
    else
    {
      // Setting old SOFT_FE version
      g_mCpcConst["NEW_SOFT_FE"] = false;
    }
    iDeviceLength = UN_CONFIG_COMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPLength(sDeviceType) + cpcConfigGenericFunctions_getConstantDPAdditionalLength(sDeviceType);
  }

  if( iConfigLength == iDeviceLength )
  {
    bHasArchive = false;
    return true;
  }
  else
  {
    if( iConfigLength == (iDeviceLength + UN_CONFIG_ADDITIONAL_ARCHIVE_LENGTH) )
    {
      bHasArchive = true;
      return true;
    }
    return false;
  }

}




/**Checks archive names for correctness.

Archives should be defined in passing importation line.
Proxy validation to unConfigGenericFunctions_checkArchive.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType DPT name
@param dsConfigs importation line, processed in the unCore
@param dsExceptionInfo exception info

*/

void cpcConfigGenericFunctions_checkArchiveConfig(string sDeviceType, dyn_string& dsConfigs, dyn_string& dsExceptionInfo)
{
  int iCommonLength;


  iCommonLength = UN_CONFIG_COMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPLength(sDeviceType);
  if( dynlen(dsConfigs) == iCommonLength + UN_CONFIG_ADDITIONAL_ARCHIVE_LENGTH + cpcConfigGenericFunctions_getConstantDPAdditionalLength(sDeviceType) )
  {
    unConfigGenericFunctions_checkArchive(dsConfigs[iCommonLength + UN_CONFIG_ADDITIONAL_ARCHIVE_BOOL],
                                          dsConfigs[iCommonLength + UN_CONFIG_ADDITIONAL_ARCHIVE_ANALOG],
                                          dsConfigs[iCommonLength + UN_CONFIG_ADDITIONAL_ARCHIVE_EVENT],
                                          dsExceptionInfo);
  }
  else
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkArchiveConfig:" + getCatStr("unGeneration", "BADLENGTH"), "");
  }

}




/**Check device address

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sFrontEndType current front-end type
@param sDeviceType device type
@param sAddressValue address reference
@param mAddressProperties checked address properties
@param bHasArchive does importation line contain bool/analog/event archives
@param dsExceptionInfo exception info

@return nothing

*/


void cpcConfigGenericFunctions_checkDeviceAddress(dyn_string dsConfigs, string sFrontEndType, string sDeviceType, string sAddressValue, mapping mAddressProperties, bool bHasArchive, dyn_string &dsExceptionInfo)
{
  float fParsedValue;
  string sOriginalAddressValue;
  dyn_string dsSplitAddress, dsExceptionInfoTmp;


  if( sAddressValue == "" )
  {
    return;
  }

  sOriginalAddressValue = sAddressValue;

  if( strpos(sAddressValue, cpcConfigGenericFunctions_sPARAM_DPE_VALUE) == 0 )
  {
    strreplace(sAddressValue, cpcConfigGenericFunctions_sPARAM_DPE_VALUE, "");
    unConfigGenericFunctions_checkFloat(sAddressValue, fParsedValue, dsExceptionInfoTmp);
    if( dynlen(dsExceptionInfoTmp) > 0 )
    {
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkDeviceAddress: bad format for set-value-instead-of-address: " + sOriginalAddressValue, "");
    }
  }
  else
  {
    switch( sFrontEndType )
    {
      case UN_PLC_DPTYPE:
        // TODO: provide unConfigGenericFunctions_checkAddressMODBUS ?
        dsSplitAddress = strsplit(sAddressValue, ".");
        if( dynlen(dsSplitAddress) > 2 )
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkDeviceAddress: bad format for UN_PLC_DPTYPE (MODBUS): " + sAddressValue, "");
        }
        unConfigGenericFunctions_checkAddresses(dsSplitAddress, dsExceptionInfo);
        break;


      case S7_PLC_DPTYPE:
        unConfigGenericFunctions_checkAddressS7(makeDynString(sAddressValue),
                                                cpcConfigGenericFunctions_translateDataTypeForS7Check(mAddressProperties["dataType"], sAddressValue),
                                                dsExceptionInfo);
        break;


      case CPC_CMW_DPTYPE:
        cpcConfigGenericFunctions_checkAddressCMW(makeDynString(sAddressValue), dsExceptionInfo);
        break;


      case CPC_OPCUA_DPTYPE:
        cpcConfigGenericFunctions_checkAddressOPCUA(makeDynString(sAddressValue), dsExceptionInfo);
        break;


      case CPC_BACNET_DPTYPE:
        // TO DO
        break;


      case CPC_DIP_DPTYPE:
        // TO DO
        break;


      case CPC_SNMP_DPTYPE:
        cpcConfigGenericFunctions_checkAddressSNMP(sAddressValue, dsExceptionInfo);
        break;


      case CPC_IEC104_DPTYPE:
        cpcConfigGenericFunctions_checkAddressIEC(sAddressValue, dsExceptionInfo);
        break;


      default:
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkDeviceAddress: unknown front end " + sFrontEndType, "");
        break;
    }
  }

}




/**Check if CMW address is correctly formered.
It's allows to have an emtpy string for address.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sAddress address string
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_checkAddressCMW(string sAddress, dyn_string &dsExceptionInfo)
{
  dyn_string dsAddressCMW, dsAddressParts;
  bool bAddressIsCorrect;
  string sType, sPollGroup, sAddressConfig;


  fwGeneral_stringToDynString(address, dsAddressCMW, CPC_CMW_ADDRESS_CONFIG_SEPARATOR);
  bAddressIsCorrect = dynlen(dsAddressCMW) == 7;
  if( !bAddressIsCorrect )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkAddressCMW: address '" + address + "' does not follow format TYPE|TRANSFORMATION_TYPE|POLL_GROUP|ADDRESS_CONFIG|FILTER_NAME|FILTER_VALUE|LLC", "");
    return;
  }

  sType          = dsAddressCMW[1];
  sPollGroup     = dsAddressCMW[3];
  sAddressConfig = dsAddressCMW[4];
  unConfigCMW_checkType(sType, sPollGroup, dsExceptionInfo);

  dsAddressParts = strsplit(sAddressConfig, "$");
  if( dynlen(dsAddressParts) != 3 )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkAddressCMW: wrong CMW address  '" + sAddressConfig + "'", "");
  }

}




/**Check 5 ranges _alert_hdl data.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param iAlarmActivePos position of alarm active field in the importation line; device's importation constant should be used.
@param diLimits positions of 4 limits - LL,L,H,HH - in the importation line; device's importation constant should be used.
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_check5RangesAlert(dyn_string dsConfigs, int iAlarmActivePos, dyn_int diLimits, dyn_string &dsExceptionInfo)
{
  float fLL, fL, fH, fHH, fOldAlarmValue, fCurrentAlarmValue;
  int iLevel, iLimitIdx;
  bool bAlarmActive, bOkRange, bError, bSms, bOldAlarmValueInit;
  dyn_string dsExceptionInfoTemp;
  string sLimit;


  cpcConfigGenericFunctions_checkBool("ALARM_ACTIVE", dsConfigs[UN_CONFIG_COMMON_LENGTH + iAlarmActivePos], bAlarmActive, dsExceptionInfo);
  if( dynlen(dsExceptionInfo) > 0 )
  {
    return;
  }


  if( dynlen(diLimits) != 4 )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_check5RangesAlert: wrong amount of alert limits: " + dynlen(diLimits) + " instead of 4.", "");
    return;
  }

  bOldAlarmValueInit = false;
  for( iLimitIdx = 1 ; iLimitIdx <= 4 ; iLimitIdx++ )
  {
    if( dsConfigs[UN_CONFIG_COMMON_LENGTH + diLimits[iLimitIdx]] != "" )
    {
      sLimit = dsConfigs[UN_CONFIG_COMMON_LENGTH + diLimits[iLimitIdx]];
      unConfigGenericFunctions_checkFloat(sLimit, fCurrentAlarmValue, dsExceptionInfoTemp);
      if( dynlen(dsExceptionInfoTemp) > 0 )
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_check5RangesAlert: wrong limit #" + (5 - iLimitIdx) + ": " + sLimit, "");
        return;
      }

      if( bOldAlarmValueInit        &&
          fOldAlarmValue >= fCurrentAlarmValue   )
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_check5RangesAlert: limits are not consistent ([limit#" + iLimitIdx + "] " + fOldAlarmValue + " >= " + fCurrentAlarmValue + " [limit#" + (iLimitIdx - 1) + "])", "");
        return;
      }
      fOldAlarmValue     = fCurrentAlarmValue;
      bOldAlarmValueInit = true;
    }
  }

}




/**Check device parameters field.

Parameters is the special field of importation line that should be presented as a KEY=VALUE comma separated list.
Parameters should contains all and the only keys that defined in device' getParamNames function.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType importation line, processed in the unCore
@param sParameters parameters value
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_checkParameters(string sDeviceType, string sParameters, dyn_string &dsExceptionInfo)
{
  dyn_string dsPossibleKeys, dsParameters, dsParameterPair, dsDeviceKeys;
  int iParametersIdx, iKeyAmount, iKeyIdx, iParametersLen, iKeyLen;
  string sKey;


  dsParameters = strsplit(sParameters, ",");

  iParametersLen = dynlen(dsParameters);
  for( iParametersIdx = 1 ; iParametersIdx <= iParametersLen ; iParametersIdx++ )
  {
    dsParameterPair = strsplit(dsParameters[iParametersIdx], "=");
    if( dynlen(dsParameterPair) == 1 )
    {
      dynAppend(dsParameterPair, "");
    }

    if( dynlen(dsParameterPair) != 2 )
    {
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkParameters:" + getCatStr("unGeneration", "BADDATA"), "");
      return;
    }
    else
    {
      dynAppend(dsPossibleKeys, dsParameterPair[1]);
    }
  }

  iKeyAmount = dynlen(dsPossibleKeys);
  if( dynUnique(dsPossibleKeys) != iKeyAmount )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkParameters: duplicate in parameters", "");
    return;
  }

  dsDeviceKeys = cpcConfigGenericFunctions_getDeviceParamNames(sDeviceType);
  iKeyLen = dynlen(dsDeviceKeys);
  for( iKeyIdx = 1 ; iKeyIdx <= iKeyLen ; iKeyIdx++ )
  {
    sKey = dsDeviceKeys[iKeyIdx];
    if( dynContains(dsPossibleKeys, sKey) < 1 )
    {
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkParameters: device should contain '" + sKey + "' parameter.", "");
      return;
    }
  }

}



/**Generic function that set device config.
Set up device's dpes using getConfig(), getParamNames() functions and importation constants defined in the device config script.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setDeviceConfig(dyn_string dsConfigs, dyn_string &dsExceptionInfo)
{
  string sDeviceType, sDpName, sFrontEndType, sCurrentAddressName, sAddress, sAddressPosition, sDeviceAlias, sDeviceLinks, sCustomFunction;
  int iDeviceLength, iParametersPos, iAddressIdx, iAddressPosition, iAcknowledgeAlarmPosition, iAlarmActivePos, iSmsCategoriesPos, iNormalPos, iLLLimitPos, iLLimitPos, iHLimitPos, iHHLimitPos, iSmsCategoryIdx, iSmsMessagesPos, iMaskEventPos, iAddressLen, iSmsCategoryLen;
  bool bHasArchive, bAcknowledgeAlarmValue, bAlarmActive, bSms, bHasSms, bMaskEventValue;
  mapping mDeviceAddresses, mCurrentAddressProps;
  dyn_string dsDeviceAddressNames, dsSmsCategories;


  sDeviceType = dsConfigs[UN_CONFIG_OBJECTGENERAL_OBJECT];
  sDpName     = dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME];

  iDeviceLength = UN_CONFIG_ALLCOMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPLength(sDeviceType) + cpcConfigGenericFunctions_getConstantDPAdditionalLength(sDeviceType);
  bHasArchive   = dynlen(dsConfigs) == iDeviceLength + UN_CONFIG_ADDITIONAL_ARCHIVE_LENGTH;
  sFrontEndType = cpcConfigGenericFunctions_getFrontEndType(dsConfigs);

  cpcConfigGenericFunctions_processArchives(UN_CONFIG_ALLCOMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPLength(sDeviceType), bHasArchive, dsConfigs);

  mDeviceAddresses     = cpcConfigGenericFunctions_getDPEConfig(sDeviceType);
  dsDeviceAddressNames = mappingKeys(mDeviceAddresses);

  iParametersPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "PARAMETERS");
  if( iParametersPos > 0 )
  {
    cpcConfigGenericFunctions_setParameters(sDeviceType, dsConfigs, dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iParametersPos], dsExceptionInfo);
    if( dynlen(dsExceptionInfo) > 0 )
    {
      return;
    }
  }

  iAddressLen = dynlen(dsDeviceAddressNames);
  for( iAddressIdx = 1 ; iAddressIdx <= iAddressLen ; iAddressIdx++ )
  {
    sCurrentAddressName  = dsDeviceAddressNames[iAddressIdx];
    mCurrentAddressProps = mDeviceAddresses[sCurrentAddressName];

    if( !cpcConfigGenericFunctions_isOldSoftFe(sFrontEndType) )
    {
      sAddress = cpcGenericDpFunctions_getDeviceProperty(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], sCurrentAddressName + "_addr", dsExceptionInfo);
      if( sAddress != "" )
      {
        cpcConfigGenericFunctions_setAddress(dsConfigs, sCurrentAddressName, mCurrentAddressProps, sAddress, sFrontEndType, dsExceptionInfo);
        //cpcGenericDpFunctions_setDeviceProperty(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], sCurrentAddressName+"_addr", "", dsExceptionInfo);
      }
      else
      {
        if( mCurrentAddressProps["dataType"] == CPC_BOOL )
        {
          sAddressPosition = cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, mCurrentAddressProps["dpe"]);
          cpcConfigGenericFunctions_setAlert(dsConfigs, sDeviceType, sCurrentAddressName, mCurrentAddressProps, sFrontEndType, mCurrentAddressProps["dpe"], dsExceptionInfo);
        }
        else
        {
          iAddressPosition = cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sCurrentAddressName);
          sAddress         = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iAddressPosition];
          cpcConfigGenericFunctions_setAddress(dsConfigs, sCurrentAddressName, mCurrentAddressProps, dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iAddressPosition], sFrontEndType, dsExceptionInfo);
        }
      }
    }


    if(  mCurrentAddressProps["isAlarm"]  &&
        !mCurrentAddressProps["stubAlarm"]  )
    {
      cpcConfigGenericFunctions_setAlertDetails(dsConfigs, sDeviceType, sCurrentAddressName, mCurrentAddressProps, sFrontEndType, mCurrentAddressProps["dpe"], dsExceptionInfo);
    }


    if( mCurrentAddressProps["hasUnit"] )
    {
      cpcConfigGenericFunctions_setUnit(dsConfigs, sDeviceType, sCurrentAddressName, mCurrentAddressProps, sFrontEndType, bHasArchive, dsExceptionInfo);
    }


    if( mCurrentAddressProps["hasFormat"] )
    {
      cpcConfigGenericFunctions_setFormat(dsConfigs, sDeviceType, sCurrentAddressName, mCurrentAddressProps, sFrontEndType, bHasArchive, dsExceptionInfo);
    }


    if( mCurrentAddressProps["hasSmooth"] )
    {
      cpcConfigGenericFunctions_setSmoothing(dsConfigs, sDeviceType, sCurrentAddressName, mCurrentAddressProps, sFrontEndType, bHasArchive, dsExceptionInfo);
    }


    if( mCurrentAddressProps["hasPvRange"] )
    {
      cpcConfigGenericFunctions_setRange(dsConfigs, sDeviceType, sCurrentAddressName, mCurrentAddressProps, sFrontEndType, bHasArchive, dsExceptionInfo);
    }


    if( mCurrentAddressProps["hasMapping"] )
    {
      cpcConfigGenericFunctions_setMapping(dsConfigs, sDeviceType, sCurrentAddressName, mCurrentAddressProps, sFrontEndType, bHasArchive, dsExceptionInfo);
    }


    if( mCurrentAddressProps["hasArchive"] )
    {
      if( strpos(sCurrentAddressName, "evStsReg") == 0 )
      {
        cpcConfigGenericFunctions_setEvArchive(dsConfigs, mCurrentAddressProps["address"], dsExceptionInfo);
      }
      else
      {
        cpcConfigGenericFunctions_setArchive(dsConfigs, sDeviceType, sCurrentAddressName, mCurrentAddressProps, sFrontEndType, bHasArchive, dsExceptionInfo);
      }
    }


    if( mCurrentAddressProps["5RangesAlerts"] )
    {
      iAcknowledgeAlarmPosition = cpcConfigGenericFunctions_getConstant(sDeviceType, "ACK_ALARM");
      cpcConfigGenericFunctions_checkBool("ACK_ALARM", dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iAcknowledgeAlarmPosition], bAcknowledgeAlarmValue, dsExceptionInfo);

      iAlarmActivePos = cpcConfigGenericFunctions_getConstant(sDeviceType, "ALARM_ACTIVE");
      cpcConfigGenericFunctions_checkBool("ALARM_ACTIVE", dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iAlarmActivePos], bAlarmActive, dsExceptionInfo);

      iSmsCategoriesPos = cpcConfigGenericFunctions_getConstant(sDeviceType, "SMS_CATEGORIES");
      bSms              = true;

      iNormalPos  = cpcConfigGenericFunctions_getConstant(sDeviceType, "NORMAL_POSITION");
      iLLLimitPos = cpcConfigGenericFunctions_getConstant(sDeviceType, "LLLIMIT");
      iLLimitPos  = cpcConfigGenericFunctions_getConstant(sDeviceType, "LLIMIT");
      iHLimitPos  = cpcConfigGenericFunctions_getConstant(sDeviceType, "HLIMIT");
      iHHLimitPos = cpcConfigGenericFunctions_getConstant(sDeviceType, "HHLIMIT");

      cpcConfigGenericFunctions_set5RangesAlert(dsConfigs,
                                                dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mCurrentAddressProps["address"],
                                                "Position Status",
                                                bAcknowledgeAlarmValue,
                                                bAlarmActive,
                                                bSms,
                                                makeDynInt(iLLLimitPos, iLLimitPos, iHLimitPos, iHHLimitPos),
                                                dsExceptionInfo);
    }

    unConfigGenericFunctions_getAliasLink(dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_ALIAS], sDeviceAlias, sDeviceLinks);

    cpcConfigGenericFunctions_setDPEAlias(sDeviceAlias, dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mCurrentAddressProps["address"], dsExceptionInfo);
    if( dynlen(dsExceptionInfo) > 0 )
    {
      DebugTN("cpcConfigGenericFunctions_setDeviceConfig() -> Error due to: " + dsExceptionInfo[2]);
      return;
    }
  }


  cpcConfigGenericFunctions_setMetainfo(sDeviceType, dsConfigs, bHasArchive, dsExceptionInfo);
  if( dynlen(dsExceptionInfo) > 0 )
  {
    return;
  }


  iSmsCategoriesPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "SMS_CATEGORIES");
  if( iSmsCategoriesPos > 0 )
  {
    dsSmsCategories = strsplit(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iSmsCategoriesPos], ",");
    iAddressLen = dynlen(dsDeviceAddressNames);
    for( iAddressIdx = 1 ; iAddressIdx <= iAddressLen ; iAddressIdx++ )
    {
      sCurrentAddressName  = dsDeviceAddressNames[iAddressIdx];
      mCurrentAddressProps = mDeviceAddresses[sCurrentAddressName];

      if( (mCurrentAddressProps["isAlarm"] && !mCurrentAddressProps["stubAlarm"]) ||
          mCurrentAddressProps["5RangesAlerts"]                                     )
      {
        unProcessAlarm_removeDpFromList(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mCurrentAddressProps["address"], dsExceptionInfo);
        if( dynlen(dsSmsCategories) > 0 )
        {
          unProcessAlarm_createCategory(dsSmsCategories, "", makeDynString(), dsExceptionInfo);
          iSmsCategoryLen = dynlen(dsSmsCategories);
          for( iSmsCategoryIdx = 1 ; iSmsCategoryIdx <= iSmsCategoryLen ; iSmsCategoryIdx++ )
          {
            bHasSms = cpcConfigGenericFunctions_alertWithMail(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], sCurrentAddressName, dsSmsCategories[iSmsCategoryIdx]);
            if( bHasSms )
            {
              unProcessAlarm_addDpToList(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mCurrentAddressProps["address"], dsSmsCategories[iSmsCategoryIdx], dsExceptionInfo);
            }
          }
        }
      }
    }
  }


  iSmsMessagesPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "SMS_MESSAGE");
  if( iSmsMessagesPos > 0 )
  {
    unGenericDpFunctions_setKeyDeviceConfiguration(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], CPC_CONFIG_SMS_MESSAGE, makeDynString(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iSmsMessagesPos]), dsExceptionInfo);
  }


  iMaskEventPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "MASKEVENT");
  if( iMaskEventPos > 0 )
  {
    cpcConfigGenericFunctions_checkBool("MASKEVENT", dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iMaskEventPos], bMaskEventValue, dsExceptionInfo);
    cpcConfigGenericFunctions_setMaskEvent(sDeviceType, dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], bMaskEventValue, dsExceptionInfo);
  }


  sCustomFunction = "CPC_";
  if( cpcConfigGenericFunctions_isOldSoftFe(sFrontEndType) )
  {
    sCustomFunction += "SOFT_FE_";
  }

  sCustomFunction += cpcConfigGenericFunctions_getShortTypeName(sDeviceType) + "Config_setCustomConfig";
  if( isFunctionDefined(sCustomFunction) )
  {
    // DebugTN("cpcConfigGenericFunctions_setDeviceConfig() -> sCustomFunction = " + sCustomFunction);
    evalScript(dsExceptionInfo, "dyn_string main(dyn_string dsConfigLine, bool bHasArchive)"              +
                                "{"                                                                       +
                                "  dyn_string dsExceptionInfo;"                                           +
                                "  " + sCustomFunction  + "(dsConfigLine, bHasArchive, dsExceptionInfo);" +
                                "  return dsExceptionInfo;"                                               +
                                "}",
                                makeDynString(),
                                dsConfigs,
                                bHasArchive);
  }

}




/**Set DPE alias to %alias%.%dpe_name% (e.g. "object1.PosSt")

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceName device name (alias)
@param sDpe full dpe name (e.g. dist_1:un-PLCATLTRT-ATLTRT-CPC_AnalogParameter-00001.ProcessInput.PosSt)
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setDPEAlias(string sDeviceName, string sDpe, dyn_string &dsExceptionInfo)
{
  dyn_string dsDpeParts;
  string sDpeName;
  int iResult;


  dsDpeParts = strsplit(sDpe, ".");
  sDpeName   = dsDpeParts[dynlen(dsDpeParts)];

  iResult = dpSetAlias(sDpe, sDeviceName + "." + sDpeName);
  if( iResult != 0 )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setDPEAlias: can't set alias for " + sDpe, "");
  }

}




/**Set description for the given element (dpe)

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sDpeName dpe name
@param mDpeProperties dpe properties
@param sFrontEndType current front-end type
@param bHasArchive does importation line contain bool/analog/event archives
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setMapping(dyn_string dsConfigs, string sDeviceType, string sDpeName, mapping mDpeProperties, string sFrontEndType, bool bHasArchive, dyn_string &dsExceptionInfo)
{
  int iDescriptionPos;
  string sDescription, sAddress;


  iDescriptionPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "MAPPING");

  sDescription = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iDescriptionPos];
  if( sDescription == "-" )
  {
    sDescription = ""; // backward compatibility after UCPC-1341
  }

  sAddress = mDpeProperties["address"];
  strreplace(sAddress, ".ProcessInput.", "");
  unGenericDpFunctions_setKeyDeviceConfiguration(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], CPC_CONFIG_MAPPING + sAddress, makeDynString(sDescription), dsExceptionInfo);

}




/**Set unit for the given element (dpe)

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sDpeName dpe name
@param mDpeProperties dpe properties
@param sFrontEndType current front-end type
@param bHasArchive does importation line contain bool/analog/event archives
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setUnit(dyn_string dsConfigs, string sDeviceType, string sDpeName, mapping mDpeProperties, string sFrontEndType, bool bHasArchive, dyn_string &dsExceptionInfo)
{
  string sUnit;
  int iUnitPos;


  iUnitPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "UNIT");
  sUnit = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iUnitPos];

  unConfigGenericFunctions_setUnit(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"], sUnit, dsExceptionInfo);

}




/**Set format for the given element (dpe)

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sDpeName dpe name
@param mDpeProperties dpe properties
@param sFrontEndType current front-end type
@param bHasArchive does importation line contain bool/analog/event archives
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setFormat(dyn_string dsConfigs, string sDeviceType, string sDpeName, mapping mDpeProperties, string sFrontEndType, bool bHasArchive, dyn_string &dsExceptionInfo)
{
  string sFormat;
  int iFormatPos;


  iFormatPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "FORMAT");
  sFormat    = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iFormatPos];

  unConfigGenericFunctions_setFormat(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"], sFormat, dsExceptionInfo);

}




/**Init range from importation line

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sDpeName dpe name
@param sRangeMax max range to set
@param sRangeMin min range to set
@param bHasArchive does importation line contain bool/analog/event archives

*/


void cpcConfigGenericFunctions_getRange(dyn_string dsConfigs, string sDeviceType, string sDpeName, string &sRangeMax, string &sRangeMin, bool bHasArchive)
{
  int iRangeMaxPos, iRangeMinPos;


  iRangeMaxPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "RANGEMAX");
  iRangeMinPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "RANGEMIN");
  if( iRangeMaxPos > 0 &&
      iRangeMinPos > 0   )
  {
    sRangeMin = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iRangeMinPos];
    sRangeMax = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iRangeMaxPos];
  }
  else
  {
    sRangeMin = "";
    sRangeMax = "";
  }

}




/**Set range for the given element (dpe)

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sDpeName dpe name
@param mDpeProperties dpe properties
@param sFrontEndType current front-end type
@param bHasArchive does importation line contain bool/analog/event archives
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setRange(dyn_string dsConfigs, string sDeviceType, string sDpeName, mapping mDpeProperties, string sFrontEndType, bool bHasArchive, dyn_string &dsExceptionInfo)
{
  float fMin, fMax;
  string sRangeMin, sRangeMax;


  cpcConfigGenericFunctions_getRange(dsConfigs, sDeviceType, sDpeName, sRangeMax, sRangeMin, bHasArchive);
  if( sRangeMin != "" &&
      sRangeMax != ""   )
  {
    // Format values
    fMin = sRangeMin;
    fMax = sRangeMax;

    if( fMin != fMax )
    {
      // if min!= max --> create pv_range config
      unConfigGenericFunctions_ScientificFormat(sRangeMin);
      unConfigGenericFunctions_ScientificFormat(sRangeMax);

      fwPvRange_set(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"],
                    sRangeMin,
                    sRangeMax,
                    UN_CONFIG_DEFAULT_RANGE_NEGATE_RANGE,
                    UN_CONFIG_DEFAULT_RANGE_IGNORE_OUTSIDE,
                    UN_CONFIG_DEFAULT_RANGE_INCLUSIVE_MIN,
                    UN_CONFIG_DEFAULT_RANGE_INCLUSIVE_MAX,
                    dsExceptionInfo);
    }
    else
    {
      // if min == max --> delete pv_range
      fwPvRange_delete(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"], dsExceptionInfo);
    }
  }
  else
  {
    fwPvRange_delete(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"], dsExceptionInfo);
  }

}




/**Set parameters for the given element (dpe)

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type
@param dsConfigs importation line, processed in the unCore
@param sParameters parameters value
@param dsExceptionInfo exception info

*/
void cpcConfigGenericFunctions_setParameters(string sDeviceType, dyn_string dsConfigs, string sParameters, dyn_string &dsExceptionInfo)
{
  dyn_string dsParameters, dsParameterParts;
  int iParametersIdx, iParametersLen;


  dpSetWait(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + UN_DEVICE_CONFIGURATION, makeDynString());

  dsParameters = strsplit(sParameters, ",");
  iParametersLen = dynlen(dsParameters);
  for( iParametersIdx = 1 ; iParametersIdx <= iParametersLen ; iParametersIdx++ )
  {
    dsParameterParts = strsplit(dsParameters[iParametersIdx], "=");
    if( dynlen(dsParameterParts) == 1 )
    {
      dynAppend(dsParameterParts, "");
    }

    cpcGenericDpFunctions_setDeviceProperty(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], dsParameterParts[1], dsParameterParts[2], dsExceptionInfo);
  }

}





/**Set meta info for the given element (dpe)

Meta info is device's heirarchy (i.e. master/parents/children), type and second alias

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type
@param dsConfigs importation line, processed in the unCore
@param bHasArchive does importation line contain bool/analog/event archives
@param dsExceptionInfo exception info

*/
void cpcConfigGenericFunctions_setMetainfo(string sDeviceType, dyn_string dsConfigs, bool bHasArchive, dyn_string &dsExceptionInfo)
{
  int iMasterNamePos, iParentsPos, iChildrenPos, iTypePos, iSecondAliasPos;


  iMasterNamePos  = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "MASTER_NAME");
  iParentsPos     = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "PARENTS");
  iChildrenPos    = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "CHILDREN");
  iTypePos        = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "TYPE");
  iSecondAliasPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "SECOND_ALIAS");

  if( iMasterNamePos > 0 )
  {
    unGenericDpFunctions_setKeyDeviceConfiguration(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], CPC_CONFIG_MASTER_NAME_KEY,  makeDynString(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iMasterNamePos]),  dsExceptionInfo);
    unGenericDpFunctions_setKeyDeviceConfiguration(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], CPC_CONFIG_PARENTS_KEY,      splitCommaSeparatedString(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iParentsPos]),     dsExceptionInfo);
    unGenericDpFunctions_setKeyDeviceConfiguration(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], CPC_CONFIG_CHILDREN_KEY,     splitCommaSeparatedString(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iChildrenPos]),    dsExceptionInfo);
    unGenericDpFunctions_setKeyDeviceConfiguration(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], CPC_CONFIG_TYPE_KEY,         splitCommaSeparatedString(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iTypePos]),        dsExceptionInfo);
    unGenericDpFunctions_setKeyDeviceConfiguration(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME], CPC_CONFIG_SECOND_ALIAS_KEY, splitCommaSeparatedString(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iSecondAliasPos]), dsExceptionInfo);
  }

}




/**Set deadband for the given element (dpe)

Deadband could be absolute, relative, old/new comparison and none.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sDpeName dpe name
@param mDpeProperties dpe properties
@param sFrontEndType current front-end type
@param bHasArchive does importation line contain bool/analog/event archives
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setSmoothing(dyn_string dsConfigs, string sDeviceType, string sDpeName, mapping mDpeProperties, string sFrontEndType, bool bHasArchive, dyn_string &dsExceptionInfo)
{
  float fDeadband;
  int iDeadbandType, iDeadbandPos, iDeadbandTypePos;


  iDeadbandPos     = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "DEADBAND");
  iDeadbandTypePos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "DEADBAND_TYPE");

  if( iDeadbandPos     > 0 &&
      iDeadbandTypePos > 0   )
  {
    fDeadband     = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iDeadbandPos];
    iDeadbandType = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iDeadbandTypePos];

    switch (iDeadbandType)
    {
      case UN_CONFIG_DEADBAND_VALUE:
        iDeadbandType = fDeadband > 0 ? DPATTR_VALUE_SMOOTH : -1;
        break;

      case UN_CONFIG_DEADBAND_RELATIF:
        iDeadbandType = fDeadband > 0 ? DPATTR_VALUE_REL_SMOOTH : -1;
        break;

      case UN_CONFIG_DEADBAND_OLD_NEW:
        iDeadbandType = DPATTR_COMPARE_OLD_NEW;
        fDeadband     = - 1;
        break;

      default:
        iDeadbandType = -1;
        break;
    }

    if( iDeadbandType == DPATTR_VALUE_SMOOTH     ||
        iDeadbandType == DPATTR_VALUE_REL_SMOOTH ||
        iDeadbandType == DPATTR_COMPARE_OLD_NEW    )
    {
      fwSmoothing_set(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"], iDeadbandType, fDeadband, UN_CONFIG_DEFAULT_SMOOTHING_TIME_FILTER, dsExceptionInfo);
    }
    else
    {
      fwSmoothing_delete(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"], dsExceptionInfo);
    }
  }

}




/**Set archiving for the given element (dpe)

For analog dpes archive parameters (including smoothing) are coded in ARCHIVE_ACTIVE and ARCHIVE_TIME_FILTER (optional) fields.
For boolean dpes it's always old/new smooth, ARCHIVE_ACTIVE is treated only as "yes" or "no".

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sDpeName dpe name
@param mDpeProperties dpe properties
@param sFrontEndType current front-end type
@param bHasArchive does importation line contain bool/analog/event archives
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setArchive(dyn_string dsConfigs, string sDeviceType, string sDpeName, mapping mDpeProperties, string sFrontEndType, bool bHasArchive, dyn_string &dsExceptionInfo)
{
  bool bArchiveActive;
  float fArchiveTimeFilter;
  int iArchiveType, iSmoothProcedure, iArchivePos, iTimeFilterPos;
  float fReturnDeadBand;
  string sArchive;


  fArchiveTimeFilter = 0;

  iArchivePos    = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "ARCHIVE_ACTIVE");
  iTimeFilterPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, sDpeName, bHasArchive, "ARCHIVE_TIME_FILTER");
  if( iArchivePos > 0 )
  {
    if( iTimeFilterPos > 0 )
    {
      fArchiveTimeFilter = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iTimeFilterPos];
    }

    unConfigGenericFunctions_getArchiving(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iArchivePos], bArchiveActive, fArchiveTimeFilter, iArchiveType, iSmoothProcedure);
    unConfigGenericFunctions_getArchivingDeadband(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iArchivePos], fReturnDeadBand);
    if( mDpeProperties["dataType"] == CPC_BOOL )
    {
      iArchiveType       = DPATTR_ARCH_PROC_SIMPLESM;
      iSmoothProcedure   = DPATTR_COMPARE_OLD_NEW;
      fReturnDeadBand    = -1;
      fArchiveTimeFilter = 0;
    }
  }
  else
  {
    bArchiveActive = false;
  }

  sArchive = mDpeProperties["dataType"] == CPC_BOOL ? dsConfigs[UN_CONFIG_OBJECTGENERAL_ARCHIVE_BOOL] : dsConfigs[UN_CONFIG_OBJECTGENERAL_ARCHIVE_ANALOG];

  if( bArchiveActive )
  {
    fwArchive_set(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"],
                  sArchive,
                  iArchiveType,
                  iSmoothProcedure,
                  fReturnDeadBand,
                  fArchiveTimeFilter,
                  dsExceptionInfo);
  }
  else
  {
    fwArchive_delete(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"], dsExceptionInfo);
  }

}




/**Set archiving for the device event dpes

The event dpe's archiving is always configured with old/new smoothing.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sAddress dpe name
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setEvArchive(dyn_string dsConfigs, string sAddress, dyn_string &dsExceptionInfo)
{
  fwArchive_set(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + sAddress,
                dsConfigs[UN_CONFIG_OBJECTGENERAL_ARCHIVE_EVENT],
                DPATTR_ARCH_PROC_SIMPLESM,
                DPATTR_COMPARE_OLD_NEW,
                0,
                0,
                dsExceptionInfo);

}




/**Set mask state for the given device

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type
@param sDeviceName device name
@param bIsMasked mask state to set
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setMaskEvent(string sDeviceType, string sDeviceName, bool bIsMasked, dyn_string &dsExceptionInfo)
{
  dyn_string dsDpes, dsDeviceAddressNames;
  mapping mDeviceAddresses;
  int iAddressIdx, iAddressLen;
  string sCurrentAddressName;


  mDeviceAddresses     = cpcConfigGenericFunctions_getDPEConfig(sDeviceType);
  dsDeviceAddressNames = mappingKeys(mDeviceAddresses);

  iAddressLen = dynlen(dsDeviceAddressNames);
  for( iAddressIdx = 1 ; iAddressIdx <= iAddressLen ; iAddressIdx++ )
  {
    sCurrentAddressName = dsDeviceAddressNames[iAddressIdx];

    if( mDeviceAddresses[sCurrentAddressName]["isInput"] )
    {
      if( mDeviceAddresses[sCurrentAddressName]["dataType"] == CPC_BOOL )
      {
        dynAppend(dsDpes, mDeviceAddresses[sCurrentAddressName]["address"]);
      }
      else
      {
        if( strpos(sCurrentAddressName, "evStsReg") == 0 )
        {
          dynAppend(dsDpes, mDeviceAddresses[sCurrentAddressName]["address"]);
        }
      }
    }
  }

  unConfigGenericFunctions_setMaskEvent(sDeviceName, bIsMasked, dsDpes, dsExceptionInfo);

}




/**Set address for the given element (dpe)

Set up periphery address for S7 or UNICOS (Schneider)

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDpeName dpe name
@param mDpeProperties dpe properties
@param sAddressValue adress position in the importation line
@param sFrontEndType current front-end type
@param dsExceptionInfo exception info

@return nothing

*/


void cpcConfigGenericFunctions_setAddress(dyn_string dsConfigs, string sDpeName, mapping mDpeProperties, string sAddressValue, string sFrontEndType, dyn_string &dsExceptionInfo)
{
  bool bPollingFrontEnd;
  int iDataType;
  string sDpe, sBitNumber, sAddressReference, sAddressMode, sPollingDp, sAddressType;


  sDpe       = dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mDpeProperties["address"];
  sBitNumber = "0";

  if( sAddressValue == "" )
  {
    fwPeriphAddress_delete(sDpe, dsExceptionInfo);
    return;
  }

  iDataType        = cpcConfigGenericFunctions_translateDataType(sFrontEndType, mDpeProperties["dataType"]);
  bPollingFrontEnd = unConfigGenericFunctions_isPollingOnly(dsConfigs[UN_CONFIG_OBJECTGENERAL_PLC_TYPE]);

  if( strpos(sAddressValue, cpcConfigGenericFunctions_sPARAM_DPE_VALUE) == 0 )
  {
    fwPeriphAddress_delete(sDpe, dsExceptionInfo);
    strreplace(sAddressValue, cpcConfigGenericFunctions_sPARAM_DPE_VALUE, "");
    dpSetWait(sDpe, sAddressValue);
  }
  else
  {
    switch( sFrontEndType )
    {
      case UN_PLC_DPTYPE:
      case S7_PLC_DPTYPE:
        if( mDpeProperties["isInput"] )
        {
          if( bPollingFrontEnd == TRUE )
          {
            sAddressMode = DPATTR_ADDR_MODE_INPUT_POLL;
          }
          else
          {
            sAddressMode = DPATTR_ADDR_MODE_INPUT_SPONT;
          }
        }
        else
        {
          sAddressMode = DPATTR_ADDR_MODE_OUTPUT;
        }


        if( sFrontEndType == UN_PLC_DPTYPE )
        {
          // For MODBUS detect bit number
          if( dynlen(strsplit(sAddressValue, ".")) > 1 )
          {
            sBitNumber = strsplit(sAddressValue, ".")[2];
          }
          else
          {
            sBitNumber = "0";
          }


          unConfigGenericFunctions_getUnicosAddressReference(makeDynString(sAddressMode,
                                                                           dsConfigs[UN_CONFIG_OBJECTGENERAL_PLC_TYPE],
                                                                           dsConfigs[UN_CONFIG_OBJECTGENERAL_PLC_NUMBER],
                                                                           strpos(sDpeName, "evStsReg") == 0,  // @TODO: is event. how to check more correct?
                                                                           sAddressValue,
                                                                           dsConfigs[UN_CONFIG_OBJECTGENERAL_EVENT16]),
                                                             sAddressReference);
        }
        else
        {
          if( sFrontEndType == S7_PLC_DPTYPE )
          {
            unConfigGenericFunctions_getUnicosAddressReferenceS7(makeDynString(sAddressMode,
                                                                               dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME],
                                                                               sAddressValue),
                                                                 sAddressReference);
          }
        }


        if( sAddressReference == "" )
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setAddress: can't create addressReference", "");
          return;
        }


        sPollingDp = "";
        if( bPollingFrontEnd )
        {
          if( sFrontEndType == UN_PLC_DPTYPE )
          {
            sPollingDp = UN_FE_POLL_DP_PREFIX + dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME];
          }
          else
          {
            if( sFrontEndType == S7_PLC_DPTYPE )
            {
              sPollingDp = "_poll_" + c_unSystemIntegrity_S7 + dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME];
            }
          }
        }


        if( sFrontEndType == UN_PLC_DPTYPE )
        {
          sAddressType = fwPeriphAddress_TYPE_MODBUS;
        }
        else
        {
          sAddressType = fwPeriphAddress_TYPE_S7;
        }

        fwPeriphAddress_set(sDpe,
                            makeDynString(sAddressType,
                                          dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER],
                                          sAddressReference,
                                          sAddressMode,
                                          iDataType,
                                          UN_CONFIG_DEFAULT_ADDRESS_ACTIVE,
                                          "",
                                          "",
                                          "",
                                          "",
                                          mDpeProperties["isInput"] ? UN_CONFIG_LOWLEVEL_ANALOG_PROCESS_INPUT : UN_CONFIG_LOWLEVEL_ANALOG_PROCESS_OUTPUT,
                                          sBitNumber, // before was hardcoded 0 - for S7 OK, but not MODBUS
                                          UN_CONFIG_DEFAULT_ADDRESS_STARTING_TIME,
                                          UN_CONFIG_DEFAULT_ADDRESS_INTERVAL_TIME,
                                          sPollingDp),
                            dsExceptionInfo,
                            FALSE,
                            FALSE);
        break;


      case CPC_CMW_DPTYPE:
        cpcConfigGenericFunctions_setCMWAddress(sDpe, sAddressValue, dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER], dsExceptionInfo);
        break;


      case CPC_OPCUA_DPTYPE:
        cpcConfigGenericFunctions_setAddressOPCUA(sDpe, sAddressValue, iDataType, dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER], dsExceptionInfo);
        break;


      case CPC_BACNET_DPTYPE:
        cpcConfigGenericFunctions_setBACnetAddress(sDpe, dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME], sAddressValue, iDataType, dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER], dsExceptionInfo);
        break;


      case CPC_DIP_DPTYPE:
        cpcConfigGenericFunctions_setDIPAddress(sDpe, dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME], sAddressValue, dsExceptionInfo);
        break;


      case CPC_SNMP_DPTYPE:
        cpcConfigGenericFunctions_setSNMPAddress(sDpe, dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME], sAddressValue, dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER], dsExceptionInfo);
        break;


      case CPC_IEC104_DPTYPE:
        cpcConfigGenericFunctions_setIEC104Address(sDpe, dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME], sAddressValue, dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER], dsExceptionInfo);
        break;


      default:
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setAddress: unknown front end " + sFrontEndType, "");
        break;
    }
  }

}




/**Set up CMW address config

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDpeName dpe name
@param sAddress address configuration in the importation line
@param sDriver driver reference
@param dsExceptionInfo exception info
*/


void cpcConfigGenericFunctions_setCMWAddress(string sDpeName, string sAddress, string sDriver, dyn_string &dsExceptionInfo)
{
  dyn_string dsFullAddressCMW, dsAddressCMW;


  fwGeneral_stringToDynString(sAddress, dsFullAddressCMW, CPC_CMW_ADDRESS_CONFIG_SEPARATOR);
  dsAddressCMW = strsplit(dsFullAddressCMW[4], "$");
  unConfigCMW_setAddress(sDpeName,
                         sDriver,
                         dsFullAddressCMW[1], // type
                         dsFullAddressCMW[2], // transformation type
                         dsFullAddressCMW[3], // poll group
                         dsFullAddressCMW[5], // filter name
                         dsFullAddressCMW[6], // filter value
                         dsAddressCMW[1],
                         dsAddressCMW[2],
                         dsAddressCMW[3],
                         dsFullAddressCMW[7], // LLC
                         dsExceptionInfo);

}




/**Check IEC104 address config

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sAddressValue, string, IEC104 address reference
@param dsExceptionInfo exception info

@return Nothing

*/


void cpcConfigGenericFunctions_checkAddressIEC(string sAddressValue, dyn_string &dsExceptionInfo)
{

  unConfigIEC104_checkAddress(sAddressValue, dsExceptionInfo);

}




cpcConfigGenericFunctions_checkAddressOPCUA(string sAddress, dyn_string &dsExceptionInfo)
{

  unConfigOPCUA_checkAddress(sAddress, dsExceptionInfo);

}




/**Check if SNMP address is correctly formered.
It's allows to have an emtpy string for address.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sAddress address string
@param dsExceptionInfo exception info

*/


cpcConfigGenericFunctions_checkAddressSNMP(string sAddress, dyn_string &dsExceptionInfo)
{
  dyn_string dsSplitAddress;
  string sTransformation, sSubindex, sPollInterval;
  int iSubIndex;


  dsSplitAddress = strsplit(sAddress, "$");
  if( dynlen(dsSplitAddress) >= 5 )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkAddressSNMP: wrong address format: " + sAddress, "");
  }
  else
  {
    unConfigSNMP_splitAddress(sAddress, sTransformation, sSubindex, sPollInterval);
    unConfigSNMP_checkAddress(sAddress, dsExceptionInfo);

    if( sSubindex != "" )
    {
      if( sscanf(sSubindex, "%d", iSubIndex) <= 0 )
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkAddressSNMP: wrong subindex:" + sSubindex, "");
      }
      else
      {
        if( iSubIndex < 0 )
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkAddressSNMP: wrong subindex value: " + iSubIndex, "");
        }
      }
    }

    if( unConfigSNMP_dataTypeStringToInt(sTransformation) == -1 )
    {
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkAddressSNMP: wrong transformation type: " + sTransformation, "");
    }
  }

}




void cpcConfigGenericFunctions_setAddressOPCUA(string sDpeName, string sAddress, int iDataType, string sDriver, dyn_string &dsExceptionInfo)
{
  int iMode;


  iMode = strpos(sDpeName, "ProcessInput") > 0 ? DPATTR_ADDR_MODE_INPUT_SPONT : DPATTR_ADDR_MODE_OUTPUT_SINGLE;
  unConfigOPCUA_setAddress(sDpeName, sAddress, mode, sDriver, iDataType, dsExceptionInfo);

}




void cpcConfigGenericFunctions_setBACnetAddress(string sDpeName, string sFrontEndName, string sAddress, int iDataType, string sDriver, dyn_string &dsExceptionInfo)
{
  int iMode;


  iMode = strpos(sDpeName, "ProcessInput") > 0 ? DPATTR_ADDR_MODE_INPUT_POLL : DPATTR_ADDR_MODE_OUTPUT;
  unConfigBACnet_setAddress(sDpeName, sFrontEndName, sAddress, mode, sDriver, iDataType, dsExceptionInfo);
}




void cpcConfigGenericFunctions_setDIPAddress(string sDpeName, string sFrontEndName, string sAddress, dyn_string &dsExceptionInfo)
{
  if( strpos(sDpeName, "ProcessInput") > 0 )
  {
    unConfigDIP_addSubscription(sDpeName, sFrontEndName, sAddress, dsExceptionInfo);
  }
  else
  {
    unConfigDIP_addPublication(sDpeName, sFrontEndName, sAddress, dsExceptionInfo, 0, true);
  }

}




void cpcConfigGenericFunctions_setSNMPAddress(string sDpeName, string sFrontEndName, string sAddressReference, string sDriverReference, dyn_string &dsExceptionInfo)
{
  int iMode, iAgentNumber, iResult, iDataType;
  string sAgentVersion, sFrontEndDpe, sAddress, sSubindex, sTransformation, sPollInterval, sPollGroup;


  iMode = strpos(sDpeName, "ProcessInput") > 0 ? DPATTR_ADDR_MODE_INPUT_POLL : DPATTR_ADDR_MODE_OUTPUT_SINGLE;

  sFrontEndDpe = getSystemName() + c_unSystemIntegrity_UNSNMP + sFrontEndName;
  iResult = dpGet(sFrontEndDpe + ".configuration.agent_number",  iAgentNumber,
                  sFrontEndDpe + ".configuration.agent_version", sAgentVersion);
  if( iResult !=0 )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkAddressSNMP: cannot retrieve values from " + sFrontEndDpe + ".configuration.agent_number and/or .configuration.agent_version", "");
    return;
  }

  sAddress        = sAddressReference;
  sSubindex       = "";
  sTransformation = "";
  sPollInterval   = "";
  unConfigSNMP_splitAddress(sAddress, sTransformation, sSubindex, sPollInterval);
  if( sSubindex == "" )
  {
    sSubindex = "0";
  }

  iDataType  = unConfigSNMP_dataTypeStringToInt(sTransformation);
  sPollGroup = unConfigSNMP_getPollGroupName(sFrontEndName, sPollInterval);
  unConfigSNMP_setAddress(sDpeName, sFrontEndName, sAddress, iMode, sDriverReference, iDataType, sSubindex, iAgentNumber, sAgentVersion, sPollGroup, dsExceptionInfo);

}




void cpcConfigGenericFunctions_setIEC104Address(string sDpeName,
                                                string sFrontEndName,
                                                string sAddressReference,
                                                string sDriver,
                                                dyn_string &dsExceptionInfo)
{
  int iMode, iDataType, iLen;
  string sReference, sCot;
  dyn_string dsReference;


  // Parsing the address reference
  dsReference = strsplit(sAddressReference, "|");
  iLen        = dynlen(dsReference);
  if( dynlen(dsReference) != iIEC104_ADDRESS_LENGTH )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setIEC104Address() -> Error extracting IEC fields from address reference, it should have " + iIEC104_ADDRESS_LENGTH + " fields and currently has: " + iLen + ", " + sAddressReference, "");
    return;
  }


  // Checking the DPE and the type of input / output
  if( strpos(sDpeName, "ProcessInput") > 0 )
  {
    iMode = DPATTR_ADDR_MODE_INPUT_SPONT;
  }
  else
  {
    // Get mode from config file
    if( (dsReference[iIEC104_ADDRESS_DIRECTION] != DPATTR_ADDR_MODE_OUTPUT)       ||
        (dsReference[iIEC104_ADDRESS_DIRECTION] != DPATTR_ADDR_MODE_OUTPUT_SINGLE)  )
    {
      iMode = DPATTR_ADDR_MODE_OUTPUT_SINGLE;
    }
  }


  // Checking the data type
  iDataType = _unConfigIEC104_convertDataToType(dsReference[iIEC104_ADDRESS_DATATYPE], dsReference[iIEC104_ADDRESS_COT], dsExceptionInfo);
  if( (iDataType == -1) ||
      (iDataType == 520)  )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setIEC104Address() -> Error translating the IEC data type: " + dsReference[iIEC104_ADDRESS_DATATYPE] +" . Selected data type is invalid: " + dsExceptionInfo[2], "");
    return;
  }


  // Creating address reference
  if( dsReference[iIEC104_ADDRESS_COT] == 1 )
  {
    sCot = "C";
  }
  sReference = "IEC_" + sFrontEndName + "-" + sCot + dsReference[iIEC104_ADDRESS_DATATYPE] + "." + dsReference[iIEC104_ADDRESS_REFERENCE];


  unConfigIEC104_setAddress(sDpeName, sReference, iMode, sDriver , iDataType, dsReference[iIEC104_ADDRESS_SUBINDEX], dsExceptionInfo);
  if( dynlen(dsExceptionInfo) > 0 )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setIEC104Address() -> Error setting the IEC address to DPE: " + sDpe + ", due to: " + dsExceptionInfo[2], "");
    return;
  }

}




/**Set alarm address for the given element (dpe)

Set up periphery address of bool dpe which is linked to StsReg for S7 or UNICOS (Schneider)

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sAlertName dpe name
@param mAlertProps dpe properties
@param sFrontEndType current front-end type
@param sDpeName name of StsReg dpe which contains the bit to link to
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setAlert(dyn_string dsConfigs, string sDeviceType, string sAlertName, mapping mAlertProperties, string sFrontEndType, string sRegisterName, dyn_string &dsExceptionInfo)
{
  bool bPollingFrontEnd;
  int iRegisterPosition;
  string sAddressReference, sAddressMode, sAddress;


  if( sFrontEndType == UN_PLC_DPTYPE ||
      sFrontEndType == S7_PLC_DPTYPE   )
  {
    bPollingFrontEnd = unConfigGenericFunctions_isPollingOnly(dsConfigs[UN_CONFIG_OBJECTGENERAL_PLC_TYPE]);
    if( bPollingFrontEnd )
    {
      sAddressMode = DPATTR_ADDR_MODE_INPUT_POLL;
    }
    else
    {
      sAddressMode = DPATTR_ADDR_MODE_INPUT_SPONT;
    }
  }


  switch( sFrontEndType )
  {
    case UN_PLC_DPTYPE:
      // Alerts for inputs only
      unConfigGenericFunctions_getUnicosAddressReference(makeDynString(sAddressMode,
                                                                       dsConfigs[UN_CONFIG_OBJECTGENERAL_PLC_TYPE],
                                                                       dsConfigs[UN_CONFIG_OBJECTGENERAL_PLC_NUMBER],
                                                                       FALSE,  // @TODO: is it correct?
                                                                       dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sRegisterName)],
                                                                       dsConfigs[UN_CONFIG_OBJECTGENERAL_EVENT16]),
                                                         sAddressReference);
      if( sAddressReference == "" )
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setAlert: can't create address reference for " + sAlertName, "");
        return;
      }

      unConfigGenericFunctions_setStsRegAlarmBitAddress(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mAlertProperties["address"],
                                                        dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER],
                                                        sAddressReference,
                                                        mAlertProperties["bitPosition"],
                                                        dsExceptionInfo);
      break;


    case S7_PLC_DPTYPE:
      unConfigGenericFunctions_getUnicosAddressReferenceS7(makeDynString(sAddressMode,
                                                                         dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME],
                                                                         dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sRegisterName)]),
                                                           sAddressReference);
      if( sAddressReference == "" )
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setAlert: can't create address reference for " + sAlertName, "");
        return;
      }

      unConfigGenericFunctions_setStsRegAlarmBitAddressS7(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mAlertProperties["address"],
                                                          dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER],
                                                          sAddressReference,
                                                          mAlertProperties["bitPosition"],
                                                          dsExceptionInfo);
      break;


    case CPC_CMW_DPTYPE:
      iRegisterPosition = cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sAlertName);
      cpcConfigGenericFunctions_setCMWAddress(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mAlertProperties["address"],
                                              dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iRegisterPosition],
                                              dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER],
                                              dsExceptionInfo);
      break;


    case CPC_OPCUA_DPTYPE:
      iRegisterPosition = cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sAlertName);
      sAddress          = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sRegisterName)];
      cpcConfigGenericFunctions_setAddressOPCUA(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mAlertProperties["address"],
                                                sAddress,
                                                OPCUA_BOOL,
                                                dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER],
                                                dsExceptionInfo);
      break;


    case CPC_IEC104_DPTYPE:
      iRegisterPosition = cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sAlertName);
      sAddress          = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + cpcConfigGenericFunctions_getConstantDPEAdress(sDeviceType, sRegisterName)];
      cpcConfigGenericFunctions_setIEC104Address(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mAlertProperties["address"],
                                                 dsConfigs[UN_CONFIG_OBJECTGENERAL_PLCNAME],
                                                 sAddress,
                                                 dsConfigs[UN_CONFIG_OBJECTGENERAL_DRIVER],
                                                 dsExceptionInfo);
      break;


    case CPC_BACNET_DPTYPE:
      // TODO
      break;


    case CPC_DIP_DPTYPE:
      // TODO
      break;


    case CPC_SNMP_DPTYPE:
      // TODO
      break;


    case SOFT_FE_DPTYPE:
      // No address
      break;


    default:
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_setAlert: unknown front end " + sFrontEndType, "");
      break;
  }

}




/**Set alarm config for the given element (dpe)

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType device type
@param sAlertName dpe name
@param mAlertProps dpe properties
@param sFrontEndType current front-end type
@param sDpeName name of StsReg dpe which contains the bit to link to
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setAlertDetails(dyn_string dsConfigs, string sDeviceType, string sAlertName, mapping mAlertProps, string sFrontEndType, string sDpeName, dyn_string &dsExceptionInfo)
{
  string bitDescription, sDomain, sNature, sDeviceAlias, sDeviceLinks, sAlertMessageOK, sAlertMessageBAD, sDescriptionText, sAlertClassPostfix, sAlertClass;
  dyn_string dsbitDescription, dsAlertTexts, dsAlertClasses;
  bool bOkRange, bAlarmExists, bAlarmActive, bSms, bAcknowledgeAlarmValue;
  int iLevel, iNormalPos, iDescriptionPos, iAcknowledgeAlarmPosition;


  unGenericDpFunctions_getBitDescription(dsConfigs[UN_CONFIG_OBJECTGENERAL_OBJECT],
                                         sDpeName,
                                         mAlertProps["bitPosition"],
                                         bitDescription);
  if( bitDescription == "" )
  {
    // Bit description was not found
    bitDescription = sAlertName;
  }

  cpcConfigGenericFunctions_createAlarmDescription(dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_DOMAIN],
                                                   dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_NATURE],
                                                   dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_ALIAS],
                                                   dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_DESCRIPTION],
                                                   bitDescription,
                                                   dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_DEFAULTPANEL],
                                                   dsbitDescription);

  // Set alarm
  iNormalPos = cpcConfigGenericFunctions_getConstant(sDeviceType, "NORMAL_POSITION");
  unConfigGenericFunctions_getNormalPosition(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iNormalPos], bAlarmExists, bAlarmActive, bOkRange, iLevel, bSms);

  iDescriptionPos = cpcConfigGenericFunctions_getConstant(sDeviceType, "ALARM_DESCRIPTION");

  if( iDescriptionPos < 1 )
  {
    sDescriptionText = cpcConfigGenericFunctions_getConstantByName("UN_CONFIG_" + strtoupper(sDeviceType), strtoupper(sAlertName) + "_ALERT_TEXT", TRUE);

    if( sDescriptionText != CPC_UNDEFINED_CONSTANT_VALUE )
    {
      sAlertMessageOK  = sDescriptionText + CPC_CONFIG_ALARM_CUSTOM_OK_POSTFIX;
      sAlertMessageBAD = sDescriptionText;
    }
    else
    {
      sAlertMessageOK  = UN_CONFIG_DEFAULT_ALERT_DIGITAL_TEXT_OK;
      sAlertMessageBAD = UN_CONFIG_DEFAULT_ALERT_DIGITAL_TEXT_BAD;
    }
  }
  else
  {
    sAlertMessageOK  = "";
    sAlertMessageBAD = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iDescriptionPos];
  }


  dsAlertTexts = bOkRange ? makeDynString(sAlertMessageBAD, sAlertMessageOK) : makeDynString(sAlertMessageOK, sAlertMessageBAD);

  sAlertClassPostfix = cpcConfigGenericFunctions_getConstantByName("UN_CONFIG_" + strtoupper(sDeviceType), strtoupper(sAlertName) + "_ALERT_CLASS_POSTFIX", TRUE);

  if( sAlertClassPostfix == CPC_UNDEFINED_CONSTANT_VALUE )
  {
    sAlertClassPostfix = "";
  }


  bAcknowledgeAlarmValue = false;
  if( sAlertName == "TStopISt"  ||
      sAlertName == "FuStopISt" ||
      sAlertName == "StartISt"  ||
      sAlertName == "AlSt"        )
  {
    bAcknowledgeAlarmValue = true;
    bSms             = false; // UCPC-2095 avoid send message for interlocks
  }
  else
  {
    bSms = true;
  }

  if( mAlertProps["hasAcknowledge"] )
  {
    iAcknowledgeAlarmPosition = cpcConfigGenericFunctions_getConstant(sDeviceType, "ACK_ALARM");
    cpcConfigGenericFunctions_checkBool("ACK_ALARM", dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + iAcknowledgeAlarmPosition], bAcknowledgeAlarmValue, dsExceptionInfo);
  }
  sAlertClass = cpcConfigGenericFunctions_getAlertClass(bAcknowledgeAlarmValue, bSms, sAlertClassPostfix);

  dsAlertClasses = bOkRange ? makeDynString(sAlertClass, "") : makeDynString("", sAlertClass);
  if( bAlarmExists )
  {
    cpcConfigGenericFunctions_setDigitalAlert(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mAlertProps["address"],
                                              dsbitDescription,
                                              dsAlertTexts,
                                              dsAlertClasses,
                                              bOkRange,
                                              bAlarmActive,
                                              bSms,
                                              dsExceptionInfo);
  }
  else
  {
    fwAlertConfig_deleteDigital(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mAlertProps["address"], dsExceptionInfo);

    unConfigGenericFunctions_setAlarmDescription(dsConfigs[UN_CONFIG_OBJECTGENERAL_DPNAME] + mAlertProps["address"],
                                                 dsbitDescription,
                                                 dsExceptionInfo);  // Set description for alert screen
  }

}




/** return true if alertName should be configured with mail handler
* returns true except there's device property %DPE%_NO_MAIL = true and %CATEGORY%_%DPE%_NO_MAIL = true
* alertName name of dpe with alert handler, for example "PosSt"

*/


bool cpcConfigGenericFunctions_alertWithMail(string sDeviceName, string sAlertName, string sCategory = "")
{
  bool bNoSmsValue;
  dyn_string dsExceptionInfo;
  string sNoSmsProperty, sNoSmsValue;


  if( "" != sCategory )
  {
    // check if for given category this alert should exist
    sNoSmsProperty = sCategory + "_" + sAlertName + NO_MAIL_POSTFIX;
    sNoSmsValue    = cpcGenericDpFunctions_getDeviceProperty(sDeviceName, sNoSmsProperty, dsExceptionInfo, "false");
    cpcConfigGenericFunctions_checkBool(sNoSmsProperty, sNoSmsValue, bNoSmsValue, dsExceptionInfo);
    if( bNoSmsValue == true )
    {
      return false;
    }
  }

  sNoSmsProperty = sAlertName + NO_MAIL_POSTFIX;
  sNoSmsValue    = cpcGenericDpFunctions_getDeviceProperty(sDeviceName, sNoSmsProperty, dsExceptionInfo, "false");
  cpcConfigGenericFunctions_checkBool(sNoSmsProperty, sNoSmsValue, bNoSmsValue, dsExceptionInfo);

  return !bNoSmsValue;

}




/**Save alerm info

Port of unConfigGenericFunctions_setDigitalAlert

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDpName datapoint name
@param dsdescription datapoint description
@param dsAlertTexts alerm messages (On and Off)
@param dsAlertClasses alert class, should exist in db
@param bOkRange deprecated
@param bActive is alarm active
@param bSms deprecated
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_setDigitalAlert(string sDpName, dyn_string dsdescription, dyn_string dsAlertTexts, dyn_string dsAlertClasses, bool bOkRange, bool bActive, bool bSms, dyn_string &dsExceptionInfo)
{
  int i, length, alertType;
  dyn_float dfLimitsGood;
  int isAlarmActive = -1;


  unConfigGenericFunctions_setAlarmDescription(sDpName, dsdescription, dsExceptionInfo);  // Set description for alert screen
  dpGet(sDpName + ":_alert_hdl.._type", alertType);
  switch( alertType )
  {
    case DPCONFIG_NONE:
      fwAlertConfig_createDigital(sDpName,
                                  dsAlertTexts,
                                  dsAlertClasses,
                                  UN_CONFIG_DEFAULT_ALERT_PANEL,
                                  UN_CONFIG_DEFAULT_ALERT_PANEL_PARAMETERS,
                                  UN_CONFIG_DEFAULT_ALERT_HELP,
                                  dsExceptionInfo);
      if( bActive )
      {
        dpSetWait(sDpName + ":_alert_hdl.._active", bActive);
      }
      break;


    case DPCONFIG_ALERT_BINARYSIGNAL:
      fwAlertConfig_modifyDigital(sDpName,
                                  dsAlertTexts,
                                  dsAlertClasses,
                                  UN_CONFIG_DEFAULT_ALERT_PANEL,
                                  UN_CONFIG_DEFAULT_ALERT_PANEL_PARAMETERS,
                                  UN_CONFIG_DEFAULT_ALERT_HELP,
                                  dsExceptionInfo);
      dpGet(sDpName + ":_alert_hdl.._active", isAlarmActive);
      if( (isAlarmActive != bActive) ||
          (isAlarmActive == -1)        )
      {
        dpSetWait(sDpName + ":_alert_hdl.._active", bActive);
      }
      break;


    case DPCONFIG_ALERT_NONBINARYSIGNAL:
    default:
      fwException_raise(dsExceptionInfo, "ERROR:", "cpcConfigGenericFunctions_setDigitalAlert :" + getCatStr("unGeneration", "BADALERT"));
      break;
  }

}




/**Create

get alarm description

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDescription description
@param dsDescription description in the format needed by unConfigGenericFunctions_setAlarmDescription

*/


cpcConfigGenericFunctions_createAlarmDescription(string sDomainConfig, string sNatureConfig, string sAlias, string sDescription, string sAlarmName, string sDefaultPanel, dyn_string &dsDescription)
{
  string sDomain, sNature, sDeviceAlias, sDeviceLinks;
  dyn_string dsDomainConfigParts, dsNatureConfigParts;


  dsDomainConfigParts = strsplit(sDomainConfig, UN_ACCESS_CONTROL_DOMAIN_SEPARATOR);
  if( dynlen(dsDomainConfigParts) == UN_CONFIG_UNICOS_DOMAIN_AND_ACCESS_CONTROL_LEN )
  {
    // access control + domain present
    sDomain = dsDomainConfigParts[2];
  }
  else
  {
    if( strpos(sDomainConfig, UN_ACCESS_CONTROL_DOMAIN_SEPARATOR) >= 0 )
    {
      // if UN_ACCESS_CONTROL_DOMAIN_SEPARATOR in sDomain config --> domain = ""
      sDomain = "";
    }
    else
    {
      sDomain = sDomainConfig;
    }
  }

  dsNatureConfigParts = strsplit(sNatureConfig, UN_ACCESS_CONTROL_DOMAIN_SEPARATOR);

  // if no ACC priv assume list of nature
  if( dynlen(dsNatureConfigParts) >= (UN_CONFIG_UNICOS_NATURE_AND_PRIV_LEN - 1) )
  {
    if( dynlen(dsNatureConfigParts) == UN_CONFIG_UNICOS_NATURE_AND_PRIV_LEN )
    {
      sNature = dsNatureConfigParts[4];
    }
    else
    {
      // dynlen = UN_CONFIG_UNICOS_NATURE_AND_PRIV_LEN -1 --> consider nature = ""
      sNature = "";
    }
  }
  else
  {
    sNature = sNatureConfig;
  }

  // alias is the deviceName only
  unConfigGenericFunctions_getAliasLink(sAlias, sDeviceAlias, sDeviceLinks);

  dsDescription [UN_CONFIG_ALARM_DOMAIN]       = sDomain;
  dsDescription [UN_CONFIG_ALARM_NATURE]       = sNature;
  dsDescription [UN_CONFIG_ALARM_ALIAS]        = sDeviceAlias;
  dsDescription [UN_CONFIG_ALARM_DESCRIPTION]  = sDescription;
  dsDescription [UN_CONFIG_ALARM_NAME]         = sAlarmName;
  dsDescription [UN_CONFIG_ALARM_DEFAULTPANEL] = sDefaultPanel;

}




/**Set 5 ranges _alert_hdl on given dpe

Port of unConfigGenericFunctions_set5RangesAlert

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDp dpe name
@param sDescription description
@param bAcknowledgeAlarmValue is alarm acknowledgeable
@param bAlarmActive is alarm active
@param bSms does alarm trigger mail/sms notifier
@param diLimits ubdex of 4 limits - LL,L,H,HH - in the importation line; device's importation constant should be used
@param dsExceptionInfo exception info

*/


void cpcConfigGenericFunctions_set5RangesAlert(dyn_string dsConfigs, string sDp, string sDescription, bool bAcknowledgeAlarmValue, bool bAlarmActive, bool bSms, dyn_int diLimits, dyn_string &dsExceptionInfo)
{
  bool bAlarmExists, bOkRange;
  dyn_string dsDescription;
  dyn_string dsAlarms, dsAlertClass;
  int iAlertType, iLevel;
  dyn_float dfLevelLimits;


  cpcConfigGenericFunctions_createAlarmDescription(dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_DOMAIN],
                                                   dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_NATURE],
                                                   dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_ALIAS],
                                                   dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_DESCRIPTION],
                                                   sDescription,
                                                   dsConfigs[UN_CONFIG_OBJECTGENERAL_LENGTH + UN_CONFIG_COMMON_DEFAULTPANEL],
                                                   dsDescription);

  bAlarmExists = dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + diLimits[1]] != "" ||
                 dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + diLimits[2]] != "" ||
                 dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + diLimits[3]] != "" ||
                 dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + diLimits[4]] != "";

  if( bAlarmExists )
  {
    dfLevelLimits = makeDynFloat();

    dpGet(sDp + ":_alert_hdl.._type", iAlertType);
    if( iAlertType != DPCONFIG_NONE )
    {
      fwAlertConfig_deleteAnalog(sDp, dsExceptionInfo);
    }

    cpcConfigGenericFunctions_get5AlertConfig(dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + diLimits[1]],
                                              dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + diLimits[2]],
                                              dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + diLimits[3]],
                                              dsConfigs[UN_CONFIG_ALLCOMMON_LENGTH + diLimits[4]],
                                              bSms,
                                              bAcknowledgeAlarmValue,
                                              dfLevelLimits,
                                              dsAlarms,
                                              dsAlertClass);

    unConfigGenericFunctions_setAnalogAlert(sDp, dsDescription, dfLevelLimits, dsAlarms, dsAlertClass, bAlarmActive, dsExceptionInfo);
  }
  else
  {
    fwAlertConfig_deleteAnalog(sDp, dsExceptionInfo);

    // even if the alarm is deleted set the description
    unConfigGenericFunctions_setAlarmDescription(sDp, dsDescription, dsExceptionInfo);
  }

}




/**Build config of 5-range alert

Port of unConfigGenericFunctions_set4LevelsAlert

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sLLParam LL threshold value
@param sLParam L threshold value
@param sHParam H threshold value
@param sHHParam HH threshold value
@param bSms does alarm trigger mail/sms notifier
@param bAcknowledgeAlarmValue is alarm acknowledgeable
@param dfLevelLimits threshold values array
@param dsLevelsAlarm threshold levels array
@param dsLevelsAlertsClass threshold classes array

*/


void cpcConfigGenericFunctions_get5AlertConfig(string sLLParam, string sLParam, string sHParam, string sHHParam, bool bSms, bool bAcknowledgeAlarmValue, dyn_float &dfLevelLimits, dyn_string &dsLevelsAlarm, dyn_string &dsLevelsAlertsClass)
{
  string sLLClass, sLClass, sHClass, sHHClass;


  sLLClass = "_cpcAnalogLL";
  sLClass  = "_cpcAnalogL";
  sHClass  = "_cpcAnalogH";
  sHHClass = "_cpcAnalogHH";

  if( bAcknowledgeAlarmValue )
  {
    sLLClass += "_Ack";
    sHHClass += "_Ack";
  }

  if( bSms )
  {
    sLLClass += "_Mail";
    sLClass  += "_Mail";
    sHClass  += "_Mail";
    sHHClass += "_Mail";
  }

  sLLClass += ".";
  sLClass  += ".";
  sHClass  += ".";
  sHHClass += ".";

  dynClear(dsLevelsAlarm);
  dynClear(dsLevelsAlertsClass);

  if( sLLParam != "" )
  {
    dynAppend(dsLevelsAlarm,       "LL");
    dynAppend(dsLevelsAlertsClass, sLLClass);
    dynAppend(dfLevelLimits,       sLLParam);
  }

  if( sLParam != "" )
  {
    dynAppend(dsLevelsAlarm,       "L");
    dynAppend(dsLevelsAlertsClass, sLClass);
    dynAppend(dfLevelLimits,       sLParam);
  }

  dynAppend (dsLevelsAlarm,       "Ok");
  dynAppend (dsLevelsAlertsClass, "");

  if( sHParam != "" )
  {
    dynAppend(dsLevelsAlarm,       "H");
    dynAppend(dsLevelsAlertsClass, sHClass);
    dynAppend(dfLevelLimits,       sHParam);
  }

  if( sHHParam != "" )
  {
    dynAppend(dsLevelsAlarm,       "HH");
    dynAppend(dsLevelsAlertsClass, sHHClass);
    dynAppend(dfLevelLimits,       sHHParam);
  }

}




/** Returns alert class according to passed params

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param bAcknowledgeAlarmValue is alarm acknowledgeable
@param bSms does alarm trigger mail/sms notifier
@param sPostFix class postfix
@param sBaseClass base class

*/


string cpcConfigGenericFunctions_getAlertClass(bool bAcknowledgeAlarmValue, bool bSms, string sPostFix = "", string sBaseClass = CPC_CONFIG_ALARM_BASE_CLASS)
{
  string sAlertClass;


  sAlertClass = sBaseClass + sPostFix;

  if( bAcknowledgeAlarmValue )
  {
    sAlertClass += CPC_CONFIG_ALARM_ACK_POSTFIX;
  }

  if( bSms )
  {
    sAlertClass += CPC_CONFIG_ALARM_MAIL_POSTFIX;
  }

  return sAlertClass;

}




/**Return the list of parameter names of device

This list is defined in DT_Config.ctl script by function called DT_getParamNames()

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type

@return device's parameter names list

*/


dyn_string cpcConfigGenericFunctions_getDeviceParamNames(string sDeviceType)
{
  string sFunction;
  dyn_string dsParameterNames;


  sFunction = "CPC_" + cpcConfigGenericFunctions_getShortTypeName(sDeviceType) + "Config_getParamNames";
  if( isFunctionDefined(sFunction) )
  {
//     DebugTN("cpcConfigGenericFunctions_getDeviceParamNames() -> sFunction = " + sFunction);
    evalScript(dsParameterNames, "dyn_string main()"                                     +
                                 "{"                                                     +
                                 "  dyn_string dsParameterNames = " + sFunction  + "();" +
                                 "  return (dsParameterNames);"                          +
                                 "}",
                                 makeDynString());
  }

  return dsParameterNames;

}




/**Returns dpes config

This mapping is defined in DT_Config.ctl script by function called DT_getConfig()

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type

@return dpes configs

*/


mapping cpcConfigGenericFunctions_getDPEConfig(string sDeviceType)
{
  mapping mAddresses;
  dyn_string dsDollars, dsKeys, dsAllProperties;
  string sFrontEndType, sDpe;
  int iKeyIdx, iPropertyIdx, iKeyLen, iPropertyLen;


  if( isDollarDefined("$sFrontEndDeviceType") )
  {
    dynAppend(dsDollars, "$sFrontEndDeviceType:" + $sFrontEndDeviceType);
  }

  evalScript(mAddresses, "mapping main()" +
                        "{"              +
                        "  return CPC_" + cpcConfigGenericFunctions_getShortTypeName(sDeviceType) + "Config_getConfig();" +
                        "}",
                        dsDollars);


  dsKeys          = mappingKeys(mAddresses);
  dsAllProperties = makeDynString("hasUnit", "hasFormat", "hasSmooth", "hasArchive", "hasPvRange", "hasMapping", "isAlarm", "stubAlarm", "hasAcknowledge", "5RangesAlerts");
  sFrontEndType   = _cpcConfigGenericFunctions_getFrontEndType();

  iKeyLen = dynlen(dsKeys);
  for( iKeyIdx = 1 ; iKeyIdx <= iKeyLen ; iKeyIdx++ )
  {
    sDpe = dsKeys[iKeyIdx];
    iPropertyLen = dynlen(dsAllProperties);
    for( iPropertyIdx = 1 ; iPropertyIdx <= iPropertyLen ; iPropertyIdx++ )
    {
      if( !mappingHasKey(mAddresses[sDpe], dsAllProperties[iPropertyIdx]) )
      {
        mAddresses[sDpe][dsAllProperties[iPropertyIdx]] = false;
      }
    }

    if( !mappingHasKey(mAddresses[sDpe], "dataType") )
    {
      mAddresses[sDpe]["dataType"] = CPC_BOOL;
    }

    mAddresses[sDpe]["isInput"] = strpos(mAddresses[sDpe]["address"], ".ProcessInput") == 0;
    if( sFrontEndType == SOFT_FE_DPTYPE )
    {
      if( mAddresses[sDpe]["hasSmooth"] )
      {
        DebugN("disable smooth options of " + sDeviceType + "-" + sDpe);
      }

      mAddresses[sDpe]["hasSmooth"] = false;
    }
  }

  return mAddresses;

}




synchronized bool cpcConfigGenericFunctions_createGlobalConst(dyn_string &dsExceptionInfo)
{
  if( addGlobal("g_mCpcConst", MAPPING_VAR) != 0 )
  {
    fwException_raise(dsExceptionInfo, "ERROR", "Error creating global variable: g_mCpcConst", "");
    return FALSE;
  }

  return TRUE;
}




synchronized bool cpcConfigGenericFunctions_AddConst(string sPrefix, string sConstant, anytype aValue)
{
  mapping mTemp;


  if( mappingHasKey(g_mCpcConst, sPrefix) == FALSE )
  {
    mTemp[sConstant]     = aValue;
    g_mCpcConst[sPrefix] = mTemp;
  }
  else
  {
    if( mappingHasKey(g_mCpcConst[sPrefix], sConstant) == FALSE )
    {
      g_mCpcConst[sPrefix][sConstant] = aValue;
    }
  }

  return TRUE;
}




/**Return the value of a device's property constant

Property's constant should be defined in DP_Config.ctl script and have following notation %FRONT_END_PREFIX%_%SHORT_DEVICE_NAME%_%PROPERTY_NAME%

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type
@param sConstantName constant name
@param bVerbose, [optional] displays if constant doesn't exist

@return device's property value

*/


int cpcConfigGenericFunctions_getConstant(string sDeviceType, string sConstantName, bool bVerbose = FALSE)
{
  bool bConstInCache;
  int iResult;
  string sName, sNamePrefix, sFrontEndType, sPrefix;
  mapping mTemp;
  dyn_string dsExceptionInfo;


  // DebugTN("cpcConfigGenericFunctions_getConstant() -> sDeviceType = " + sDeviceType + ", sConstantName = " + sConstantName);

  bConstInCache = FALSE;

  if( globalExists("g_mCpcConst") == FALSE )
  {
    // Create global
    if( cpcConfigGenericFunctions_createGlobalConst(dsExceptionInfo) == FALSE )
    {
      DebugTN("cpcConfigGenericFunctions_getConstant() -> Error, gathering constant for device type: " + sDeviceType + ", constant name: " + sConstantName + ", due to problems checking global variable g_mCpcConst: " + dsExceptionInfo[2]);
      return -1;
    }
  }

  // DebugTN("cpcConfigGenericFunctions_getConstant() -> check g_mCpcConst = ", g_mCpcConst);

  sFrontEndType = _cpcConfigGenericFunctions_getFrontEndType();
  sPrefix       = cpcConfigGenericFunctions_getConstantPrefix(sFrontEndType);
  sNamePrefix   = sPrefix + strtoupper(sDeviceType);

  // Check if constant exist in the cache
  if( mappingHasKey(g_mCpcConst, sNamePrefix) == TRUE )
  {
    if( mappingHasKey(g_mCpcConst[sNamePrefix], sConstantName) == TRUE )
    {
      // Return the constant in the cache
      // DebugTN("cpcConfigGenericFunctions_getConstant() -> constant available in cache, value = ", g_mCpcConst[sNamePrefix][sConstantName] );

      bConstInCache = TRUE;
      iResult       = g_mCpcConst[sNamePrefix][sConstantName];
    }
  }


  if( bConstInCache == FALSE )
  {
    // We need to query the constant and cache it
    sName = sNamePrefix + "_" + sConstantName;
    evalScript(iResult, "int main()"                    +
                        "{"                             +
                        "  int iResult = -1;"           +
                        "  try"                         +
                        "  {"                           +
                        "    iResult = " + sName + ";"  +
                        "  }"                           +
                        "  catch"                       +
                        "  {"                           +
                        "    iResult = -1;"             +
                        "  }"                           +
                        "  return iResult;"             +
                        "}",
                        makeDynString());


    // Update the constant in the cache
    if( mappingHasKey(g_mCpcConst, sNamePrefix) == FALSE )
    {
      // NamePrefix item doesn't exist
      mTemp[sConstantName]     = iResult;
      g_mCpcConst[sNamePrefix] = mTemp;
    }
    else
    {
      if( mappingHasKey(g_mCpcConst[sNamePrefix], sConstantName) )
      {
        if( g_mCpcConst[sNamePrefix][sConstantName] != iResult )
        {
          DebugTN("cpcConfigGenericFunctions_getConstant() -> Error, constant already exists and new value found, please contact icecontrols.support@cern.ch -> Previous = " + g_mCpcConst[sNamePrefix][sConstantName] + " -> New = " + iResult);
          g_mCpcConst[sNamePrefix][sConstantName] = iResult;
        }
        // else  // Constant is the same than in the cache and doesn't need to be updated
      }
      else
      {
        // NamePrefix and next mapping already exists
        g_mCpcConst[sNamePrefix][sConstantName] = iResult;
      }
    }
  }


  // DebugTN("cpcConfigGenericFunctions_getConstant() -> iResult = ", iResult);


  if( bVerbose && iResult == -1 )
  {
    DebugTN("cpcConfigGenericFunctions_getConstant() -> Error, Can't find required constant " + sName);
  }

  return iResult;

}




/**Returns address position in the importation line for passed DPE name

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type
@param sAddress DPE name

@return address position in the import line

*/


int cpcConfigGenericFunctions_getConstantDPEAdress(string sDeviceType, string sAddress)
{
  int iPosition;
  string sConstantPrefix, sFrontEndType, sConstantName;


  // DebugTN("cpcConfigGenericFunctions_getConstantDPEAdress() -> sDeviceType = " + sDeviceType + ", sAddress = " + sAddress);

  sFrontEndType        = _cpcConfigGenericFunctions_getFrontEndType();
  sConstantPrefix      = cpcConfigGenericFunctions_getConstantPrefix(sFrontEndType);
  iPosition            = cpcConfigGenericFunctions_getConstantByName(sConstantPrefix + strtoupper(sDeviceType), "ADDRESS_" + strtoupper(sAddress));

  //   DebugTN("cpcConfigGenericFunctions_getConstantDPEAdress() -> iPosition = " + iPosition);

  return (iPosition);

}




/**Returns config position for given device type, periphery address name and property name.

Function make a lookup for constant UN_CONFIG_[DEVICE]_[ADDRESS_NAME]_[PROP]. if not found then try constant UN_CONFIG_[DEVICE]_[PROP]

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType, device type
@param sAddressName, DPE name
@param bHasArchive, does importation line contain bool/analog/event archives
@param sPropertyName, DPE's property name

*/


int cpcConfigGenericFunctions_getConstantDPEProperty(string sDeviceType, string sAddressName, bool bHasArchive, string sPropertyName)
{
  int iPosition;
  string sName, sPrefix, sFrontEndType;


  // DebugTN("cpcConfigGenericFunctions_getConstantDPEProperty() -> sDeviceType = " + sDeviceType + ", sAddressName = " + sAddressName + ", bHasArchive = " + bHasArchive + ", sPropertypName = " + sPropertyName);

  iPosition     = 0;
  sFrontEndType = _cpcConfigGenericFunctions_getFrontEndType();
  sPrefix       = cpcConfigGenericFunctions_getConstantPrefix(sFrontEndType);
  sName         = sPrefix + strtoupper(sDeviceType);
  // DebugTN("cpcConfigGenericFunctions_getConstantDPEProperty() -> sName = " + sName);


  if( strlen(sAddressName) > 0 )
  {
    iPosition = cpcConfigGenericFunctions_getConstantByName(sName, "ADDITIONAL_" + strtoupper(sAddressName) + "_" + strtoupper(sPropertyName));
  }


  if( iPosition < 1 )
  {
    iPosition = cpcConfigGenericFunctions_getConstantByName(sName, "ADDITIONAL_" + strtoupper(sPropertyName));
  }


  if( iPosition > 0 )
  {
    // Calculate full position
    iPosition = iPosition + cpcConfigGenericFunctions_getConstantDPLength(sDeviceType);
    if( bHasArchive )
    {
      iPosition = iPosition + UN_CONFIG_ADDITIONAL_ARCHIVE_LENGTH;
    }

  }


  if( iPosition < 1 && strlen(sAddressName) > 0 )
  {
    iPosition = cpcConfigGenericFunctions_getConstantByName(sName, strtoupper(sAddressName) + "_" + strtoupper(sPropertyName));
  }


  if( iPosition < 1 )
  {
    iPosition = cpcConfigGenericFunctions_getConstantByName(sName, strtoupper(sPropertyName));
  }

  return iPosition;

}




/**Returns a length of importation line

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type

@return length of the importation line

*/


int cpcConfigGenericFunctions_getConstantDPLength(string sDeviceType)
{
  int iLength;


  iLength = (int) cpcConfigGenericFunctions_getConstant(strtoupper(sDeviceType) , "LENGTH", TRUE);

  return iLength;

}




/**Returns an additional length of importation line

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type

@return additional lenght of importation line

*/


int cpcConfigGenericFunctions_getConstantDPAdditionalLength(string sDeviceType)
{
  int iLength;


  iLength = (int) cpcConfigGenericFunctions_getConstant(sDeviceType, "ADDITIONAL_LENGTH");
  if( iLength < 0 )
  {
    iLength = 0;
  }

  return iLength;

}




/**Returns a value of constant by name

Returns -1 if value is not defined.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sName, constant name

@return constant's value

*/


anytype cpcConfigGenericFunctions_getConstantByName(string sPrefix, string sConstant, bool bSpecialConstant = FALSE)
{
  bool bConstInCache;
  string sName;
  anytype aResult, aReturn;
  mapping mTemp;
  dyn_string dsExceptionInfo;


  // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> sPrefix = " + sPrefix + ", sConstant = " + sConstant);

  bConstInCache = FALSE;

  if( globalExists("g_mCpcConst") == FALSE )
  {
    // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> creating global g_mCpcConst");
    if( cpcConfigGenericFunctions_createGlobalConst(dsExceptionInfo) == FALSE )
    {
      DebugTN("cpcConfigGenericFunctions_getConstantByName() -> Error, gathering constant by name for device type: " + sDeviceType + ", constant name: " + sConstant + ", due to problems checking global variable g_mCpcConst: " + dsExceptionInfo[2]);
      return -1;
    }
  }


  // Check if constant exist in the cache
  if( mappingHasKey(g_mCpcConst, sPrefix) == TRUE )
  {
    if( mappingHasKey(g_mCpcConst[sPrefix], sConstant) == TRUE )
    {
      // Return the constant in the cache
      // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> constant available in cache, value = ", g_mCpcConst[sPrefix][sConstant]);
      // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> constant available in cache, g_mCpcConst[sPrefix] = ", g_mCpcConst[sPrefix]);

      bConstInCache = TRUE;
      aReturn       = g_mCpcConst[sPrefix][sConstant];
    }
  }


  if( bConstInCache == FALSE )
  {
    // We need to query the constant and cache it
    if( sPrefix == "UNICOS" )
    {
      sName = sConstant;
    }
    else
    {
      sName = sPrefix + "_" + sConstant;
    }
    // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> sName query = ", sName);

    evalScript(aResult, "anytype main()"               +
                        "{"                            +
                        "  anytype aResult;"           +
                        "  try"                        +
                        "  {"                          +
                        "    aResult = " + sName + ";" +
                        "  }"                          +
                        "  catch"                      +
                        "  {"                          +
                        "    aResult = -1;"            +
                        "  }"                          +
                        "  return aResult;"
                        "}",
                        makeDynString());

    // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> aResult = ", aResult);

    if( (bSpecialConstant == TRUE) && (aResult == -1) )
    {
      // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> Change to CPC_UNDEFINED_CONSTANT_VALUE: " + CPC_UNDEFINED_CONSTANT_VALUE);
      aReturn = (string) CPC_UNDEFINED_CONSTANT_VALUE;
    }
    else
    {
      aReturn = aResult;
    }
    // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> aReturn = ", aReturn);


    // Update the constant in the cache
    if( mappingHasKey(g_mCpcConst, sPrefix) == FALSE )
    {
      // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> Update constant in cache: prefix doesn't exist in cache = " + sPrefix );

      // NamePrefix item doesn't exist
      mTemp[sConstant]     = aReturn;
      g_mCpcConst[sPrefix] = mTemp;
    }
    else
    {
      // NamePrefix and next mapping already exists
      // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> Update constant in cache: prefix already exists in cache = " + sPrefix + " -> ", g_mCpcConst[sPrefix]);

      if( mappingHasKey(g_mCpcConst[sPrefix], sConstant) )
      {
        if( g_mCpcConst[sPrefix][sConstant] != aReturn )
        {
          DebugTN("cpcConfigGenericFunctions_getConstantByName() -> Error, constant already exists and new value found, please contact icecontrols.support@cern.ch -> Previous = " + g_mCpcConst[sPrefix][sConstant] + " -> New = " + aReturn);
          g_mCpcConst[sPrefix][sConstant] = aReturn;
        }
        // else  // Constant is the same than in the cache and doesn't need to be updated

      }
      else
      {
        // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> Update constant in cache[" + sPrefix + "]: Constant didn't exist: " + sConstant);
        g_mCpcConst[sPrefix][sConstant] = aReturn;
      }
    }
  }
  // DebugTN("cpcConfigGenericFunctions_getConstantByName() -> exit aReturn = ", aReturn);


  return aReturn;

}




/**Returns a string value of constant by name

Returns CPC_UNDEFINED_CONSTANT_VALUE if value is not defined.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sName constant name

@return constant's value

*/


string cpcConfigGenericFunctions_getStringConstantByName(string sName)
{
  string sResult;


  //DebugTN("cpcConfigGenericFunctions_getStringConstantByName() -> sName = " + sName);
  evalScript(sResult, "anytype main()"                              +
                      "{"                                           +
                      "  string sResult = \"\";"                    +
                      "  try"                                       +
                      "  {"                                         +
                      "    sResult = " + sName + ";"                +
                      "  }"                                         +
                      "  catch"                                     +
                      "  {"                                         +
                      "    sResult = CPC_UNDEFINED_CONSTANT_VALUE;" +
                      "  }"                                         +
                      "  return sResult;"                           +
                      "}",
                      makeDynString());


  //DebugTN("cpcConfigGenericFunctions_getStringConstantByName() -> sResult = " + sResult);

  return sResult;

}




/**Returns a device type name without CPC_ prefix

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type
@return DT name without CPC_ prefix
*/


string cpcConfigGenericFunctions_getShortTypeName(string sDeviceType)
{
  string sDeviceTypeCopy;


  sDeviceTypeCopy = sDeviceType;
  strreplace(sDeviceTypeCopy, "CPC_", "");

  return sDeviceTypeCopy;

}




/**Returns a position constants prefix

Can be SOFT_FE for SOFT Front-End or UN_CONFIG for others.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sFrontEndType, Front end type
@return front-end related prefix
*/


string cpcConfigGenericFunctions_getConstantPrefix(string sFrontEndType)
{
  string sPrefix;


  sPrefix = "UN_CONFIG_";
  if( cpcConfigGenericFunctions_isOldSoftFe(_cpcConfigGenericFunctions_getFrontEndType()) )
  {
    sPrefix = sPrefix + "SOFT_FE_";
  }


  return sPrefix;

}




/**Returns selected front-end type

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@return sDeviceType device type

*/


string _cpcConfigGenericFunctions_getFrontEndType()
{
  // TODO: refactor me. animation can call this...
  // TODO: we should remove this function. It creates a wrong abstraction that front end type is always available, like a global value. it doesn't correspond the reality
  // the problem is that in device check it's the only way to get front end type.
  string sFrontEndType;


  if( isFunctionDefined("shapeExists") && shapeExists("TextFieldPlcType") )
  {
    sFrontEndType = unConfigGenericFunctions_extractFeProtocol(TextFieldPlcType.text);
    if( sFrontEndType == UN_CONFIG_PROTOCOL_MODBUS ||
        sFrontEndType == UN_CONFIG_PREMIUM         ||
        sFrontEndType == UN_CONFIG_QUANTUM         ||
        sFrontEndType == UN_CONFIG_UNITY             )
    {
      sFrontEndType = UN_PLC_DPTYPE;
    }
    else
    {
      if( sFrontEndType == UN_CONFIG_PROTOCOL_S7 ||
          sFrontEndType == UN_CONFIG_S7_400      ||
          sFrontEndType == UN_CONFIG_S7_300        )
      {
        sFrontEndType = S7_PLC_DPTYPE;
      }
    }
  }
  else
    if( isDollarDefined("$sFrontEndDeviceType") )
    {
      sFrontEndType = $sFrontEndDeviceType;
    }
    else
    {
      DebugN("_cpcConfigGenericFunctions_getFrontEndType: can't detect front-end type");
    }

  return sFrontEndType;

}




string cpcConfigGenericFunctions_getFrontEndType(dyn_string dsConfig)
{

  return unImportDevice_getFrontEndDeviceType(dsConfig[UN_CONFIG_OBJECTGENERAL_PLC_TYPE]);

}




/**Convert CPC data-type representation to the front-end specific

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sFrontEndType front-end to translate to
@param commonDataType CPC data type

@return int, fromt end data type

*/


int cpcConfigGenericFunctions_translateDataType(string sFrontEndType, int iCommonDataType)
{
  switch( sFrontEndType )
  {
    case UN_PLC_DPTYPE:
      switch( iCommonDataType )
      {
        case CPC_INT16:
          return PVSS_MODBUS_INT16;
          break;

        case CPC_INT32:
          return PVSS_MODBUS_INT32;
          break;

        case CPC_UINT16:
          return PVSS_MODBUS_UINT16;
          break;

        case CPC_BOOL:
          return PVSS_MODBUS_BOOL;
          break;

        case CPC_FLOAT:
          return PVSS_MODBUS_FLOAT;
          break;

        default:
          break;
      }
      break;


    case CPC_OPCUA_DPTYPE:
      switch( iCommonDataType )
      {
        case CPC_INT16:
          return OPCUA_INT16;
          break;

        case CPC_INT32:
          return OPCUA_INT32;
          break;

        case CPC_UINT16:
          return OPCUA_UINT16;
          break;

        case CPC_BOOL:
          return OPCUA_BOOL;
          break;

        case CPC_FLOAT:
          return OPCUA_FLOAT;
          break;

        default:
          break;
      }
      break;


    case CPC_BACNET_DPTYPE:
      // TODO
      switch( iCommonDataType )
      {
        case CPC_INT16:
          return 800;
          break;

        case CPC_INT32:
          return 800;
          break;

        case CPC_UINT16:
          return 800;
          break;

        case CPC_BOOL:
          return 800;
          break;

        case CPC_FLOAT:
          return 800;
          break;

        default:
          break;
      }
      break;


    case CPC_SNMP_DPTYPE:
      switch( iCommonDataType )
      {
        case CPC_INT16:
          return 661;
          break;

        case CPC_INT32:
          return 661;
          break;

        case CPC_UINT16:
          return 661;
          break;

        case CPC_BOOL:
          return 661;
          break;

        case CPC_FLOAT:
          return 661;
          break;

        default:
          break;
      }
      break;


    case CPC_IEC104_DPTYPE:
      // To be done
      switch( iCommonDataType )
      {
        case CPC_INT16:
          return PVSS_S7_INT16;
          break;

        case CPC_INT32:
          return PVSS_S7_INT32;
          break;

        case CPC_UINT16:
          return PVSS_S7_UINT16;
          break;

        case CPC_BOOL:
          return PVSS_S7_BOOL;
          break;

        case CPC_FLOAT:
          return PVSS_S7_FLOAT;
          break;

        default:
          break;
      }
      break;


    default:
      switch( iCommonDataType )
      {
        case CPC_INT16:
          return PVSS_S7_INT16;
          break;

        case CPC_INT32:
          return PVSS_S7_INT32;
          break;

        case CPC_UINT16:
          return PVSS_S7_UINT16;
          break;

        case CPC_BOOL:
          return PVSS_S7_BOOL;
          break;

        case CPC_FLOAT:
          return PVSS_S7_FLOAT;
          break;

        default:
          break;

      }
      break;
  }

}//  cpcConfigGenericFunctions_translateDataType()




/**Convert CPC data-type representation to the S7 check data type

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param commonDataType CPC data type
@param sAddressValue S7 Address to be checked

@return S7 data type check

*/
int cpcConfigGenericFunctions_translateDataTypeForS7Check(int commonDataType, string sAddressValue)
{
  // First try to get the type from the address
  string sAddressPart1, sAddressPart2, sAddressPart3;
  int iAddressPart1, iAddressPart2, iAddressPart3;


  sscanf(sAddressValue, "%[^0-9]%d%[^0-9]%d%[^0-9]%d", sAddressPart1, iAddressPart1, sAddressPart2, iAddressPart2, sAddressPart3, iAddressPart3);
  if( sAddressPart2 == "." + UN_S7_FORMAT_WORD )
  {
    return UN_CONFIG_S7_PLC_DATATYPE_WORD;
  }
  else
  {
    if( sAddressPart2 == "." + UN_S7_FORMAT_BIT )
    {
      return UN_CONFIG_S7_PLC_DATATYPE_BOOL;
    }
    else
    {
      if( sAddressPart2 == "." + UN_S7_FORMAT_BYTE )
      {
        return UN_CONFIG_S7_PLC_DATATYPE_BYTE;
      }
    }
  }

  // If the type is DOUBLE or not found, then get it from the commonDataType
  switch( commonDataType )
  {
    case CPC_INT16:
    case CPC_UINT16:
      return UN_CONFIG_S7_PLC_DATATYPE_WORD;
      break;

    case CPC_INT32:
      return UN_CONFIG_S7_PLC_DATATYPE_DWORD;
      break;

    case CPC_BOOL:
      return UN_CONFIG_S7_PLC_DATATYPE_BOOL; //@TODO: verify with reality
      break;

    case CPC_FLOAT:
      return UN_CONFIG_S7_PLC_DATATYPE_FLOAT;
      break;
  }

}




/**Split comma-separated string and trim each element

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@todo move to generic place. reuse in other places. merge with cpcGenericObject_StringToMapping

@param sData comma-separated string

@return array of trimmed elements

*/


dyn_string splitCommaSeparatedString(string sData)
{
  dyn_string dsResult;
  int iDataIdx, iDataLen;


  dsResult = strsplit(sData, ",");

  iDataLen = dynlen(dsResult);
  for( int iDataIdx = 1 ; iDataIdx <= iDataLen ; iDataIdx++ )
  {
    dsResult[iDataIdx] = strrtrim(strltrim(dsResult[iDataIdx]));
  }

  return dsResult;

}




/**Gets archiving configs from object and puts them to the common archiving configs.

@todo move to the unicos core

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param iToSkip Num elements in dsConfigs to skip before common part
@param bHasArchive is archived defined in the importation line
@param dsConfigs importation line, processed in the unCore

*/


void cpcConfigGenericFunctions_processArchives(int iToSkip, bool bHasArchive, dyn_string& dsConfigs)
{
  if( bHasArchive )
  {
    if( dsConfigs[iToSkip + UN_CONFIG_ADDITIONAL_ARCHIVE_BOOL] != "" )
    {
      dsConfigs[UN_CONFIG_OBJECTGENERAL_ARCHIVE_BOOL] = dsConfigs[iToSkip + UN_CONFIG_ADDITIONAL_ARCHIVE_BOOL];
    }

    if( dsConfigs[iToSkip + UN_CONFIG_ADDITIONAL_ARCHIVE_ANALOG] != "" )
    {
      dsConfigs[UN_CONFIG_OBJECTGENERAL_ARCHIVE_ANALOG] = dsConfigs[iToSkip + UN_CONFIG_ADDITIONAL_ARCHIVE_ANALOG];
    }

    if( dsConfigs[iToSkip + UN_CONFIG_ADDITIONAL_ARCHIVE_EVENT] != "" )
    {
      dsConfigs[UN_CONFIG_OBJECTGENERAL_ARCHIVE_EVENT] = dsConfigs[iToSkip + UN_CONFIG_ADDITIONAL_ARCHIVE_EVENT];
    }
  }

}




/**Gets additional parameters configs from object and puts them to a mapping.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsConfigs importation line, processed in the unCore
@param sDeviceType Device type name
@param bHasArchive is archived defined in the importation line

*/


mapping cpcConfigGenericFunctions_getParametersMapping(dyn_string dsConfigs, string sDeviceType, bool bHasArchive)
{
  int iParametersPos, iParametersIdx, iParameterLen;
  dyn_string dsParameters, dsEntry;
  mapping mParameters;
  string sKey, sValue;


  iParametersPos = cpcConfigGenericFunctions_getConstantDPEProperty(sDeviceType, "", bHasArchive, "PARAMETERS");
  dsParameters   = strsplit(dsConfigs[UN_CONFIG_COMMON_LENGTH + iParametersPos], ",");

  iParameterLen = dynlen(dsParameters);
  for( iParametersIdx = 1 ; iParametersIdx <= iParameterLen ; iParametersIdx++ )
  {
    dsEntry = strsplit(dsParameters[iParametersIdx], "=");
    sKey    = dsEntry[1];
    sValue  = "";
    if( dynlen(dsEntry) > 1 )
    {
      sValue = dsEntry[2];
    }

    mParameters[sKey] = sValue;
  }
  return mParameters;

}




/**Checks message convertion parameter

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sConversion string with message convertion configuration
@param dsExceptionInfo for errors
*/


void cpcConfigGenericFunctions_checkMsgConv(string sConversion, dyn_string &dsExceptionInfo)
{
  dyn_string dsConversion;
  string sType;
  float fValue;


  if( sConversion != "" )
  {
    dsConversion = strsplit(sConversion, "|");
    sType        = strrtrim(strltrim(strtolower(dsConversion[1])));
    if( sType == "linear" )
    {
      if( dynlen(dsConversion) != 3 )
      {
        fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkMsgConv: wrong set of parameters for linear transformation: " + getCatStr("unGeneration", "BADINPUTS"), "");
        return;
      }
      else
      {
        unConfigGenericFunctions_checkFloat(dsConversion[2], fValue, dsExceptionInfo);
        if( dynlen(dsExceptionInfo) > 0 )
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkMsgConv: wrong float value: " + dsConversion[2] + ", " + getCatStr("unGeneration", "BADINPUTS"), "");
          return;
        }

        unConfigGenericFunctions_checkFloat(dsConversion[3], fValue, dsExceptionInfo);
        if( dynlen(dsExceptionInfo) > 0 )
        {
          fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkMsgConv: wrong float value: " + dsConversion[3] + ", " + getCatStr("unGeneration", "BADINPUTS"), "");
          return;
        }
      }
    }
    else
    {
      fwException_raise(dsExceptionInfo, "ERROR", "cpcConfigGenericFunctions_checkMsgConv: wrong type of transformation: " + sType + ", " + getCatStr("unGeneration", "BADINPUTS"), "");
      return;
    }
  }

}




/** Sets _msg_conv dpe config based on given parameter string

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sConversion string with message convertion configuration
@param dpes list of dpes to e configured
@param dsExceptionInfo for errors

*/


void cpcConfigGenericFunctions_setMsgConv(string sConversion, dyn_string dsDpes, dyn_string &dsExceptionInfo)
{
  dyn_string dsConversion;
  string sType;
  int iDpeIdx, iDpeLen;


  if( sConversion != "" )
  {
    dsConversion = strsplit(sConversion, "|");
    sType        = strrtrim(strltrim(strtolower(dsConversion[1])));

    iDpeLen = dynlen(dsDpes);
    for( iDpeIdx = 1 ; iDpeIdx <= iDpeLen ; iDpeIdx++ )
    {
      if( sType == "linear" )
      {
        fwConfigConversion_set(dsDpes[iDpeIdx], DPCONFIG_CONVERSION_RAW_TO_ENG_MAIN, DPDETAIL_CONV_POLY, 1, makeDynFloat(dsConversion[2], dsConversion[3]), dsExceptionInfo);
      }
    }
  }

}




/** Adds system name to the dp if not present

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sTextBox contains the dp to be modified in its text attribute
*/


void cpcConfigGenericFunctions_eventCommand(string sTextBox)
{
  string sDpName;
  dyn_string ds;
  int pos;


  //get the datapoint
  getValue(sTextBox, "text", sDpName);

  // if no sys name is defined then should take the local
  if( strrtrim(strltrim(sDpName)) != "" )
  {
    if( strpos(sDpName, ":") <0 )
    {
      //no check on the system is done
      //because this is to allow the user to develop in his own lab without having to be connected to the hardware

      // if there is a $ given assume that this is a dollar parameter
      if( strpos(sDpName, "$") < 0 )
      {
        sDpName = getSystemName() + sDpName;
      }
    }
    setValue(sTextBox, "text", sDpName);
  }

}


/** Checks if the old Soft FE is used

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sPlcType FrontEnd type
*/
bool cpcConfigGenericFunctions_isOldSoftFe(string sFrontEndType)
{
    return (sFrontEndType == SOFT_FE_DPTYPE && (!mappingHasKey(g_mCpcConst, "NEW_SOFT_FE") || !g_mCpcConst["NEW_SOFT_FE"]));
}




