#uses "cpcExportGenericFunctionsDeprecated.ctl"




/**@file

@brief This library contains generic exportation functions.

Normally a function in this library is designed to accept currentDP and dsDpParameters (array representation of pvss importation line) with two first parameters.
Other parameters may be accepted in order to specify a particular element in device to process.
Function adds certain parameters to the end of dsDpParameters as a result of its work.

@author Nikolay KULMAN (IT/CO)
@author Alexey Merezhin (EN-ICE-PLC)

@copyright
  &copy;Copyright CERN 2013 - all rights reserved

@par Creation Date
  06/04/2006

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI, CTRL

*/

/**Add DPE address to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param sDeviceType device type
@param sDpeName dpe name (without path)
@param bIsInput should be true if it's in ProcessInput and false if it's in ProcessOutput
@param dsDpParameters object's output
@return nothing
*/

void cpcExportGenericFunctions_processAddress(string sDp, string sDeviceType, string sDpeName, bool bIsInput, dyn_string &dsDpParameters)
{
  int iUpdateTime;
  string sDrvIdent, sAddr, sPrefix, sDpe, sFunction;
  mixed mValue;


  // Get the Dpe name
  if( bIsInput == TRUE )
  {
    sPrefix = ".ProcessInput.";
  }
  else
  {
    sPrefix = ".ProcessOutput.";
  }
  sDpe = sDp + sPrefix + sDpeName;

  dpGet(sDpe + ":_address.._drv_ident" , sDrvIdent);

  switch( sDrvIdent )
  {
    case "IEC":
      sDrvIdent = "IEC104";
      break;

    case "":
      break;

    default:
      break;
  }

  sFunction = "DRV_" + sDrvIdent + "_convertToUnicosAddress";
  if( isFunctionDefined(sFunction) )
  {
    evalScript(sAddr, "string main(string sDeviceDpeName, string sDeviceType)"    +
                      "{"                                                         +
                      "  string sAddr;"                                           +
                      "  sAddr = " + sFunction + "(sDeviceDpeName, sDeviceType);" +
                      "  return sAddr;"                                           +
                      "}",
               makeDynString(),
               sDpe,
               sDeviceType);
  }
  else
  {
    // Address not set
    dpGet(sDpe,                        mValue,
          sDpe + ":_original.._stime", iUpdateTime);
    if( iUpdateTime == 0 )
    {
      sAddr = UN_CONFIG_EXPORT_NONE;
    }
    else
    {
      sAddr = cpcConfigGenericFunctions_sPARAM_DPE_VALUE + mValue;
    }
  }

  dynAppend(dsDpParameters, sAddr);

}//  cpcExportGenericFunctions_processAddress()





/**
Purpose: function to retrieve CMW address for given Data Point

Parameters:
  sDeviceDpeName: string, input, dpe name for which address should be formatted
  sDeviceType: string, input, device type name

Usage: External function

PVSS manager usage: Ctrl, NG, NV

@reviewed 2018-07-24 @whitelisted{Callback}
*/


string DRV_CMW_convertToUnicosAddress(string sDeviceDpeName, string sDeviceType)
{
  dyn_string dsAddressConfig;


  unConfigCMW_getAddressConfig(sDeviceDpeName, dsAddressConfig);

  if( dsAddressConfig[4] == "" )
  {
    // no address config
    return "";
  }
  else
  {
    return dsAddressConfig[1] + "|" +                                                       // type
           dsAddressConfig[2] + "|" +                                                       // transformation
           dsAddressConfig[3] + "|" +                                                       // poll group, interval
           dsAddressConfig[4] + "$" + dsAddressConfig[5] + "$" + dsAddressConfig[6] + "|" + // device$property$tag
           dsAddressConfig[7] + "|" +                                                       // filter name
           dsAddressConfig[8] + "|" +                                                       // filter value
           dsAddressConfig[9];                                                              // LLC
  }

}//  DRV_CMW_convertToUnicosAddress()




/**Add parameters field to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param dsExceptionInfo exception info
*/


void cpcExportGenericFunctions_getParameters(string sDp, dyn_string &dsDpParameters, dyn_string &dsExceptionInfo)
{
  int iKeyIndex, iLen;
  string sKey, sValue;
  dyn_string dsParams, dsKey;
  dyn_dyn_string ddsConfig;


  _unGenericDpFunctions_getConfiguration(sDp, UN_DEVICE_CONFIGURATION, dsKey, ddsConfig, dsExceptionInfo);

  iLen = dynlen(dsKey);
  for( iKeyIndex = 1 ; iKeyIndex <= iLen ; iKeyIndex++ )
  {
    sKey = dsKey[iKeyIndex];

    if( strpos(sKey, UN_CPC_DEVICE_PROPERTY_PREFIX) == 0 )
    {
      strreplace(sKey, UN_CPC_DEVICE_PROPERTY_PREFIX, "");
      sValue = ddsConfig[iKeyIndex];
      dynAppend(dsParams, sKey + "=" + sValue);
    }
  }

  dynAppend(dsDpParameters, dynJoin(dsParams, ","));

}// cpcExportGenericFunctions_getParameters()





/**Add meta-info fields to the device's output

Meta info is device's heirarchy (i.e. master/parents/children), type and second alias

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param dsExceptionInfo exception info
*/


void cpcExportGenericFunctions_getMetainfo(string sDp, dyn_string &dsDpParameters, dyn_string &dsExceptionInfo)
{
  dyn_string dsMaster, dsParents, dsChildren, dsType, dsSecondAlias;


  unGenericDpFunctions_getKeyDeviceConfiguration(sDp, CPC_CONFIG_MASTER_NAME_KEY,  dsMaster,      dsExceptionInfo);
  unGenericDpFunctions_getKeyDeviceConfiguration(sDp, CPC_CONFIG_PARENTS_KEY,      dsParents,     dsExceptionInfo);
  unGenericDpFunctions_getKeyDeviceConfiguration(sDp, CPC_CONFIG_CHILDREN_KEY,     dsChildren,    dsExceptionInfo);
  unGenericDpFunctions_getKeyDeviceConfiguration(sDp, CPC_CONFIG_TYPE_KEY,         dsType,        dsExceptionInfo);
  unGenericDpFunctions_getKeyDeviceConfiguration(sDp, CPC_CONFIG_SECOND_ALIAS_KEY, dsSecondAlias, dsExceptionInfo);

  dynAppend(dsDpParameters, dynJoin(dsMaster,      ","));
  dynAppend(dsDpParameters, dynJoin(dsParents,     ","));
  dynAppend(dsDpParameters, dynJoin(dsChildren,    ","));
  dynAppend(dsDpParameters, dynJoin(dsType,        ","));
  dynAppend(dsDpParameters, dynJoin(dsSecondAlias, ","));

}// cpcExportGenericFunctions_getMetainfo()





/**Add unit to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param sDpe name of unit's dpe, ProcessInput.PosSt by default
*/


void cpcExportGenericFunctions_getUnit(string sDp, dyn_string &dsDpParameters, string sDpe = ".ProcessInput.PosSt")
{
  string sUnit;


  sUnit = dpGetUnit(sDp + sDpe);
  dynAppend(dsDpParameters, sUnit);

}// cpcExportGenericFunctions_getUnit()





/**Add format to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param sDpe name of format's dpe, ProcessInput.PosSt by default
*/

void cpcExportGenericFunctions_getFormat(string sDp, dyn_string &dsDpParameters, string sDpe = ".ProcessInput.PosSt")
{
  string sFormat;
  dyn_string dsExceptionInfo;

  sFormat = unExportDevice_getFormat(sDp, sDpe, dsExceptionInfo);
  dynAppend(dsDpParameters, sFormat);

}// cpcExportGenericFunctions_getFormat()





/**Add format without log_scale flag to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param sDpe name of format's dpe, ProcessInput.PosSt by default
*/

void cpcExportGenericFunctions_getSimpleFormat(string sDp, dyn_string &dsDpParameters, string sDpe = ".ProcessInput.PosSt")
{
  string sFormat;


  sFormat = unExportDevice_convertFormat(dpGetFormat(sDp + sDpe), "#");
  dynAppend(dsDpParameters, sFormat);

}//  cpcExportGenericFunctions_getSimpleFormat()





/**Add range fields (max, min) to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param dsExceptionInfo exception info
@param sDpe name of range's dpe, ProcessInput.PosSt by default
*/


void cpcExportGenericFunctions_getRange(string sDp, dyn_string &dsDpParameters, dyn_string &dsExceptionInfo, string sDpe = ".ProcessInput.PosSt")
{
  bool bDoesExist, bNegateRange, bIgnoreOutside, bInclusiveMin, bInclusiveMax;
  float fMinValue, fMaxValue;


  fwPvRange_get(sDp + sDpe, bDoesExist, fMinValue, fMaxValue, bNegateRange, bIgnoreOutside, bInclusiveMin, bInclusiveMax, dsExceptionInfo);
  dynAppend(dsDpParameters, unExportDevice_floatFormat(fMaxValue));
  dynAppend(dsDpParameters, unExportDevice_floatFormat(fMinValue));

}//  cpcExportGenericFunctions_getRange()





/**Add out range fields (max, min) to the device's output

Out range stands for the scaled range.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param dsExceptionInfo exception info
@param sDpe name of out range's dpe, ProcessInput.PosSt by default
*/

void cpcExportGenericFunctions_getOutRange(string sDp, dyn_string &dsDpParameters, dyn_string &dsExceptionInfo, string sDpe = ".ProcessInput.PosSt")
{
  bool bDoesExist, bNegateRange, bIgnoreOutside, bInclusiveMin, bInclusiveMax, bConfigExists;
  int iConversionType, iOrder;
  float fRangeMin, fRangeMax, fOutRangeMax, fOutRangeMin;
  dyn_float dfArguments;


  fwPvRange_get(sDp + sDpe, bDoesExist, fRangeMin, fRangeMax, bNegateRange, bIgnoreOutside, bInclusiveMin, bInclusiveMax, dsExceptionInfo);
  fwConfigConversion_get(sDp + sDpe, bConfigExists, DPCONFIG_CONVERSION_RAW_TO_ENG_MAIN, iConversionType, iOrder, dfArguments, dsExceptionInfo);
  fOutRangeMax = dfArguments[2] * fRangeMax + dfArguments[1];
  fOutRangeMin = dfArguments[2] * fRangeMin + dfArguments[1];

  dynAppend(dsDpParameters, unExportDevice_floatFormat(fOutRangeMax));
  dynAppend(dsDpParameters, unExportDevice_floatFormat(fOutRangeMin));

}//  cpcExportGenericFunctions_getOutRange()





/**Add deadband fields (deadband and deadband type) to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param dsExceptionInfo exception info
@param sDpe name of deadband's dpe, ProcessInput.PosSt by default
*/


void cpcExportGenericFunctions_getDeadband(string sDp, dyn_string &dsDpParameters, dyn_string &dsExceptionInfo, string sDpe = ".ProcessInput.PosSt")
{
  bool bIsSmoothDefined;
  int iSmoothProcedure, iDeadbandType;
  float fDeadband, fTimeInterval;


  unExportDevice_fwSmoothing_getTemp(sDp + sDpe, bIsSmoothDefined, iSmoothProcedure, fDeadband, fTimeInterval);
  if( bIsSmoothDefined )
  {
    switch( iSmoothProcedure )
    {
      case DPATTR_VALUE_REL_SMOOTH:
        iDeadbandType = UN_CONFIG_DEADBAND_RELATIF;
        break;

      case DPATTR_VALUE_SMOOTH:
        iDeadbandType = UN_CONFIG_DEADBAND_VALUE;
        break;

      case DPATTR_COMPARE_OLD_NEW:
        iDeadbandType = UN_CONFIG_DEADBAND_OLD_NEW;
        fDeadband     = 0;
        break;

      default:
        iDeadbandType = DPCONFIG_NONE;
        fDeadband     = 0;
        break;
    }
  }
  else
  {
    iDeadbandType = DPCONFIG_NONE;
    fDeadband     = 0;
  }

  dynAppend(dsDpParameters, fDeadband);
  dynAppend(dsDpParameters, iDeadbandType);

}//  cpcExportGenericFunctions_getDeadband()





/**Add archive fields (archive active and archive time filter) to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param sDpe name of archive's dpe, ProcessInput.PosSt by default
*/

void cpcExportGenericFunctions_getArchive(string sDp, dyn_string &dsDpParameters, string sDpe = ".ProcessInput.PosSt")
{
  float fArchiveTimeFilter;
  string sArchiveActive;


  unExportDevice_getArchiveParametersDpe(sDp + sDpe, sArchiveActive, fArchiveTimeFilter);
  dynAppend(dsDpParameters, sArchiveActive);
  if( fArchiveTimeFilter < 0 )
  {
    dynAppend(dsDpParameters, "");
  }
  else
  {
    dynAppend(dsDpParameters, fArchiveTimeFilter);
  }

}//  cpcExportGenericFunctions_getArchive()





/**Add isDeviceMasked value to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
*/

void cpcExportGenericFunctions_getMaskEvent(string sDp, dyn_string &dsDpParameters)
{
  int iMask;
  string sMask;


  unGenericDpFunctions_getMaskEvent(sDp, iMask);
  sprintf(sMask, "%d", iMask);
  dynAppend(dsDpParameters, sMask);

}//  cpcExportGenericFunctions_getMaskEvent()





/**Add default value to the device's output

Function stands for fetch online value of ProcessOutput DPE

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param sDefaultElementName dpe name, "DfltVal" by default
*/

void cpcExportGenericFunctions_getDefaultValue(string sDp, dyn_string &dsDpParameters, string sDefaultElementName = "DfltVal")
{
  float fDefaultValue;


  dpGet(sDp + ".ProcessOutput." + sDefaultElementName + ":_online.._value", fDefaultValue);
  dynAppend(dsDpParameters, fDefaultValue);

}//  cpcExportGenericFunctions_getDefaultValue()





/**Add default value casted to bool to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@see cpcExportGenericFunctions_getDefaultValue

@param sDp current dp name
@param dsDpParameters object's output
@param sDefaultElementName dpe name, "DfltVal" by default
*/

void cpcExportGenericFunctions_getBoolDefaultValue(string sDp, dyn_string &dsDpParameters, string sDefaultElementName = "DfltVal")
{
  bool bDefaultValue;


  dpGet(sDp + ".ProcessOutput." + sDefaultElementName + ":_online.._value", bDefaultValue);
  dynAppend(dsDpParameters, bDefaultValue);

}//  cpcExportGenericFunctions_getBoolDefaultValue()





/**Add DPE values to the device's output

Function fetches the online value of DPE

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp Current DP name
@param sElementName The DPE name
@param dsDpParameters A dyn_string to which the value of the DPE is appended; the output of the function
@param sPrintFormat Optional sprintf() format used to print the DPE value; default is "", i.e. no custom formatting
@param sConfigDetailAttrib Optional config, detail and attribute of the DPE; default is ":_online.._value", replicating the default dpGet() behaviour in WccOA 3.15
*/
void cpcExportGenericFunctions_getDPEValue(string sDp, string sElementName, dyn_string &dsDpParameters, string sPrintFormat="", string sConfigDetailAttrib=":_online.._value")
{
  anytype aReturnValue;  // holds value to be returned
  string sReturnType;    // the type of the returned valued, used for debugging only
  string sFullNameDpe;   // the complete name of the DPE whose values is to be returned


  sFullNameDpe = sDp + sElementName + sConfigDetailAttrib;
  sReturnType  = dpAttributeType(sFullNameDpe);
  dpGet(sFullNameDpe, aReturnValue);


  if( sPrintFormat == "" )
  {
    dynAppend(dsDpParameters, aReturnValue);
    DebugFN(2, "Exporting online value of type " + sReturnType + ": " + sFullNameDpe + " = " + aReturnValue);
  }
  else
  {
    string sFormatted;

    sprintf(sFormatted, sPrintFormat, aReturnValue);
    DebugFN(2, "Exporting online value of type " + sReturnType + ": " + sFullNameDpe + " = " + sFormatted);
    dynAppend(dsDpParameters, sFormatted);
  }

}//  cpcExportGenericFunctions_getDPEValue()





/**Add PID name to the device's output

Proxy to unExportDevice_getPIDName

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
*/

void cpcExportGenericFunctions_getPIDName(string sDp, dyn_string &dsDpParameters)
{
  string sPIDName;


  unExportDevice_getPIDName(sDp, sPIDName);
  dynAppend(dsDpParameters, sPIDName);

}//  cpcExportGenericFunctions_getPIDName()





/**Add device config value for given name to the device's output

Proxy to unExportDevice_getPCOName

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param sKey key in device configuration
@param dsExceptionInfo exception info
*/

void cpcExportGenericFunctions_getKeyDeviceConfiguration(string sDp, dyn_string &dsDpParameters, string sKey, dyn_string &dsExceptionInfo)
{
  string sValue;


  unGenericDpFunctions_getKeyDeviceConfiguration(sDp, sKey, sValue, dsExceptionInfo);
  dynAppend(dsDpParameters, sValue);

}//  cpcExportGenericFunctions_getKeyDeviceConfiguration()





/**Add digital normal position to the device's output

Port of unExportDevice_getDigitalNormalPositionDPE.

Use full path in alarm list (dsDpElements), for example, there's a list of interlocks for a field object:

    makeDynString(".ProcessInput.StartISt", ".ProcessInput.TStopISt", ".ProcessInput.FuStopISt")

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDP current dp name
@param sObject device type; don't used
@param dsDpElements an array of alarm dpes
@param dsDpParameters object's output
*/

void cpcExportGenericFunctions_getDigitalNormalPosition(string sDP, string sObject, dyn_string dsDpElements, dyn_string &dsDpParameters)
{
  // sObject is not used at the moment, but kept as unicos functions need it. you can pass "".
  bool bAlarmMask, bOkRange, bBoolAlert;
  int iAlertType, iRes, iNormalPosition, iDpIndex, iLen;
  string sDataPoint, sAlertClass1, sAlertClass2;
  dyn_int diAlertClass;


  iNormalPosition = -1;
  iLen            = dynlen(dsDpElements);

  for( iDpIndex = 1 ; iDpIndex <= iLen ; iDpIndex++ )
  {
    iRes = dpGet(sDP + dsDpElements[iDpIndex] + ":_alert_hdl.._type", iAlertType);
    dynAppend(diAlertClass, iAlertType);
  }


  if( dynUnique(diAlertClass) != 1 )
  {
    iAlertType = DPCONFIG_NONE;
  }
  else
  {
    iAlertType = diAlertClass[1];
  }


  if( iAlertType == DPCONFIG_NONE )
  {
    iNormalPosition = 2;
  }
  else
  {
    sDataPoint = sDP + dsDpElements[1];
    bBoolAlert = ((iRes >= 0) && (iAlertType == DPCONFIG_ALERT_BINARYSIGNAL));
    if( bBoolAlert )
    {
      dpGet(sDataPoint + ":_alert_hdl.._ok_range", bOkRange);
    }

    dpGet(sDataPoint + ":_alert_hdl.._active",  bAlarmMask,
          sDataPoint + ":_alert_hdl.1._class",  sAlertClass1,
          sDataPoint + ":_alert_hdl.2._class",  sAlertClass2);

    if( bAlarmMask )
    {
      // Alarm unmasked
      iNormalPosition = bOkRange ? 1 : 0;
    }
    else
    {
      // Alarm masked
      iNormalPosition = bOkRange ? 4 : 3;
    }
  }
  dynAppend(dsDpParameters, iNormalPosition);

}//  cpcExportGenericFunctions_getDigitalNormalPosition()





/**Add PCO parameters to the device's output

Proxy to unExportDevice_getPCOParameters.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDP current dp name
@param dsDpParameters object's output
*/

void cpcExportGenericFunctions_getPCOParameters(string sDP, dyn_string &dsDpParameters)
{
  int iPCOParamIndex, iLen;
  dyn_string dsPCOParameters;


  unExportDevice_getPCOParameters(sDP, dsPCOParameters);
  iLen = dynlen(dsPCOParameters);

  for( iPCOParamIndex = 1 ; iPCOParamIndex <= iLen ; iPCOParamIndex++ )
  {
    dynAppend(dsDpParameters, dsPCOParameters[iPCOParamIndex]);
  }

}//  cpcExportGenericFunctions_getPCOParameters()





/**Add autonomous property to devices

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param dsDpList list of dp names
@param dsExceptionInfo exception info
@param sDpe name of autonomous's dpe, ProcessInput.PosSt by default
*/

void cpcExportGenericFunctions_setAutonomous(dyn_string dsDpList, dyn_string &dsExceptionInfo, string sDpe = ".ProcessInput.PosSt")
{
  int i, iType, iResult;
  string sCurrentDp, sValue;
  dyn_string dsDpParameters;


  for( i = 1; i <= dynlen(dsDpList); i++ )
  {
    sCurrentDp = dsDpList[i];
    cpcExportGenericFunctions_getParameters(sCurrentDp, dsDpParameters, dsExceptionInfo);
    if( !dynContains(dsDpParameters, "AUTONOMOUS=1") &&
        !dynContains(dsDpParameters, "AUTONOMOUS=0")   )
    {
      iResult = dpGet(sCurrentDp + sDpe + ":_dp_fct.._type", iType);
      sValue = iType == DPCONFIG_DP_FUNCTION ? "1" : "0";
      cpcGenericDpFunctions_setDeviceProperty(sCurrentDp, "AUTONOMOUS", sValue, dsExceptionInfo);
    }
  }
}




/**Add archive name of given dpes to the device's output

Funtion collect archive name for each of dpes.
If archive is the same for everyone, it use this archive.
Else it use empty string.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param sDefaultValue value that used if archive is not configured for dpe
@param dsDpes list of dpes that supposed to have same archive
@param dsDpParameters object's output
*/

void cpcExportGenericFunctions_getArchiveNameForDpes(string sDp, string sDefaultValue, dyn_string dsDpes, dyn_string &dsDpParameters)
{
  int iLoop, iLen;
  string sArchiveName, sArchiveClass, sName;
  dyn_string dsCheck;


  if( dynlen(dsDpes) == 0 )
  {
    sArchiveName = sDefaultValue;
  }
  else
  {
    iLen = dynlen(dsDpes);
    for( iLoop = 1; iLoop <= iLen ; iLoop++ )
    {
      dpGet(sDp + "." + dsDpes[iLoop] + ":_archive.1._class", sArchiveClass);

      if( dpExists(sArchiveClass + ".general.arName") )
      {
        dpGet(sArchiveClass + ".general.arName:_online.._value", sName);
      }
      else
      {
        if( dpSubStr(sArchiveClass, DPSUB_DP_EL_CONF_DET_ATT) == "_EVENT" )
        {
          sName = "RDB-99) EVENT";
        }
        else
        {
          sName = sDefaultValue;
        }
      }

      dynAppend(dsCheck, sName);
    }

    dynUnique(dsCheck);
    if( dynContains(dsCheck, "") > 0 )
    {
      dynRemove(dsCheck, dynContains(dsCheck, ""));
    }

    if( dynlen(dsCheck) == 1 )
    {
      //check if same archive for same type
      sArchiveName = dsCheck[1];
    }
    else
    {
      sArchiveName = "";
    }
  }

  dynAppend(dsDpParameters, sArchiveName);

}//  cpcExportGenericFunctions_getArchiveNameForDpes()





/**Return list of dpes of given deviceType that have hasArchive property

Funtion collect dpes' names that are archived

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDeviceType device type name
@param iArchiveType type of archive: 1 (UN_CONFIG_ADDITIONAL_ARCHIVE_BOOL) for BOOLEAN, 2 (UN_CONFIG_ADDITIONAL_ARCHIVE_ANALOG) for ANALOG and 3 (UN_CONFIG_ADDITIONAL_ARCHIVE_EVENT) for EVENT

@return list of dpes
*/

dyn_string cpcExportGenericFunctions_getArchivedDpes(string sDeviceType, int iArchiveType)
{
  int iKeyIndex, iLen;
  string sDpe;
  dyn_string dsArchivedDpes, dsKeys;
  mapping mAddresses;


  mAddresses = callFunction("CPC_" + cpcConfigGenericFunctions_getShortTypeName(sDeviceType) + "Config_getConfig");
  dsKeys     = mappingKeys(mAddresses);
  iLen       = dynlen(dsKeys);

  for( iKeyIndex = 1; iKeyIndex <= iLen ; iKeyIndex++ )
  {
    sDpe = dsKeys[iKeyIndex];
    if( mappingHasKey(mAddresses[sDpe], "hasArchive") )
    {
      switch( iArchiveType )
      {
        case UN_CONFIG_ADDITIONAL_ARCHIVE_BOOL:
          if( (mappingHasKey(mAddresses[sDpe], "bitPosition")) ||
              (mappingHasKey(mAddresses[sDpe], "dataType") && mAddresses[sDpe]["dataType"] == CPC_BOOL) )
          {
            dynAppend(dsArchivedDpes, strltrim(mAddresses[sDpe]["address"], "."));
          }
          break;

        case UN_CONFIG_ADDITIONAL_ARCHIVE_ANALOG:
          if( (mappingHasKey(mAddresses[sDpe], "dataType")) &&
              (dynContains(makeDynInt(CPC_FLOAT, CPC_INT16, CPC_UINT16), mAddresses[sDpe]["dataType"]) > 0) )
          {
            dynAppend(dsArchivedDpes, strltrim(mAddresses[sDpe]["address"], "."));
          }
          break;

        case UN_CONFIG_ADDITIONAL_ARCHIVE_EVENT:
          if( mappingHasKey(mAddresses[sDpe], "dataType") &&
              mAddresses[sDpe]["dataType"] == CPC_INT32     )
          {
            dynAppend(dsArchivedDpes, strltrim(mAddresses[sDpe]["address"], "."));
          }
          break;
      }
    }
  }

  return dsArchivedDpes;

}//  cpcExportGenericFunctions_getArchivedDpes()





/**Add alarm categories assigned to device to the device's output

Proxy for unProcessAlarm_getCategory.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpes dyn_string with the DPEs
@param dsDpParameters object's output
@param dsExceptionInfo exception info
*/


void cpcExportGenericFunctions_getCategories(string sDp, dyn_string dsDpes, dyn_string &dsDpParameters, dyn_string &dsExceptionInfo)
{
  int iDpeIndex, iDpCategoriesIndex, iLen;
  string sCategories;
  dyn_string dsDpCategories;
  dyn_dyn_string ddsCategory;


  iLen = dynlen(dsDpes);
  for( iDpeIndex = 1 ; iDpeIndex <= iLen ; iDpeIndex++ )
  {
    dsDpes[iDpeIndex] = sDp + dsDpes[iDpeIndex];
  }

  unProcessAlarm_getCategory(dsDpes, ddsCategory, dsExceptionInfo);

  if( dynlen(ddsCategory) > 0 )
  {
    sCategories    = "";
    dsDpCategories = ddsCategory[1];

    iLen = dynlen(dsDpCategories);
    for( iDpCategoriesIndex = 1 ; iDpCategoriesIndex <= iLen ; iDpCategoriesIndex++ )
    {
      if( sCategories != "" )
      {
        sCategories += ",";
      }
      sCategories += dsDpCategories[iDpCategoriesIndex];
    }

    dynAppend(dsDpParameters, sCategories);

  }
  else
  {
    dynAppend(dsDpParameters, "");
  }

}//  cpcExportGenericFunctions_getCategories()





/**Returns if dpe's alarm acknowledgeable

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param sDpe name of acknowledgeable's dpe, ProcessInput.PosSt by default
*/

string cpcExportGenericFunctions_getAcknowledgeAlarmValue(string sDp, string sDpe = ".ProcessInput.PosSt")
{
  int iRes, iAlertType;
  string sClassName;


  iRes = dpGet(sDp + sDpe + ":_alert_hdl.._type", iAlertType);
  if( (iRes < 0)                    ||
      (iAlertType == DPCONFIG_NONE)   )
  {
    return "N";
  }
  else
  {
    dpGet(sDp + sDpe + ":_alert_hdl.._class", sClassName);
    if( strpos(sClassName, CPC_CONFIG_ALARM_ACK_POSTFIX) >= 0 )
    {
      return "Y";
    }
    else
    {
      return "N";
    }
  }

}//  cpcExportGenericFunctions_getAcknowledgeAlarmValue()





/**Add acknowledgeable value to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@todo unify format: rename alarmDP to dpe and move to the end of parameters

@param sDp current dp name
@param sAlarmDp name of acknowledgeable's dpe
@param dsDpParameters object's output
@param dsExceptionInfo exception info
*/

void cpcExportGenericFunctions_getAcknowledgeAlarm(string sDp, string sAlarmDp, dyn_string &dsDpParameters, dyn_string &dsExceptionInfo)
{
  string sIsAck;


  sIsAck = cpcExportGenericFunctions_getAcknowledgeAlarmValue(sDp, sAlarmDp);
  dynAppend(dsDpParameters, sIsAck);

}//  cpcExportGenericFunctions_getAcknowledgeAlarm()





/**Add alarm's threshold values to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
*/

void cpcExportGenericFunctions_get5AlertLimits(string sDp, dyn_string &dsDpParameters)
{
  int iRangesNumber, iRangeIndex;
  float fValue;
  string sRangeClass;
  dyn_string dsLimits;


  dsLimits = makeDynString("", "", "", "");
  dpGet(sDp + ".ProcessInput.PosSt:_alert_hdl.._num_ranges", iRangesNumber);

  for( iRangeIndex = 1 ; iRangeIndex <= iRangesNumber ; iRangeIndex++ )
  {
    sRangeClass = "#";

    dpGet(sDp + ".ProcessInput.PosSt:_alert_hdl." + iRangeIndex + "._class", sRangeClass);

    if( strpos(sRangeClass, "_cpcAnalogHH") > 0 )
    {
      dpGet(sDp + ".ProcessInput.PosSt:_alert_hdl." + iRangeIndex + "._l_limit", fValue);
      dsLimits[1] = fValue;
    }
    else
    {
      if( strpos(sRangeClass, "_cpcAnalogH") > 0 )
      {
        dpGet(sDp + ".ProcessInput.PosSt:_alert_hdl." + iRangeIndex + "._l_limit", fValue);
        dsLimits[2] = fValue;
      }
      else
      {
        if( strpos(sRangeClass, "_cpcAnalogLL") > 0 )
        {
          dpGet(sDp + ".ProcessInput.PosSt:_alert_hdl." + iRangeIndex + "._u_limit", fValue);
          dsLimits[4] = fValue;
        }
        else
        {
          if( strpos(sRangeClass, "_cpcAnalogL") > 0 )
          {
            dpGet(sDp + ".ProcessInput.PosSt:_alert_hdl." + iRangeIndex + "._u_limit", fValue);
            dsLimits[3] = fValue;
          }
        }
      }
    }
  }

  dynAppend(dsDpParameters, dsLimits[1]);
  dynAppend(dsDpParameters, dsLimits[2]);
  dynAppend(dsDpParameters, dsLimits[3]);
  dynAppend(dsDpParameters, dsLimits[4]);

}//  cpcExportGenericFunctions_get5AlertLimits()





/**Fetch classes of 5-range alert

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@todo return dsClasses instead

@param sDp current dp name
@param dsClasses output
@param dsExceptionInfo exception info
@param sDpe name of dpe with 5-range alert, ProcessInput.PosSt by default
*/


void cpcExportGenericFunctions_get5AlertClasses(string sDp, dyn_string &dsClasses, dyn_string &dsExceptionInfo, string sDpe = ".ProcessInput.PosSt")
{
  int iRangesNumber, iRangesNumberIndex;
  string sRangeClass;


  dynClear(dsClasses);


  dpGet(sDp + sDpe + ":_alert_hdl.._num_ranges", iRangesNumber);
  for( iRangesNumberIndex = 1 ; iRangesNumberIndex <= iRangesNumber ; iRangesNumberIndex++ )
  {
    dpGet(sDp + sDpe + ":_alert_hdl." + iRangesNumberIndex + "._class", sRangeClass);
    if( sRangeClass != "" )
    {
      dynAppend(dsClasses, sRangeClass);
    }
  }

}//  cpcExportGenericFunctions_get5AlertClasses()





/**Return is 5-range alert acknowledgeable

If at least one of thresholds contains CPC_CONFIG_ALARM_ACK_POSTFIX in the alert class, alert is acknowledgeable
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsExceptionInfo exception info
*/

string cpcExportGenericFunctions_get5AlertAcknowledgeAlarmValue(string sDp, dyn_string &dsExceptionInfo)
{
  int iClassesIndex, iLen;
  dyn_string dsClasses;



  cpcExportGenericFunctions_get5AlertClasses(sDp, dsClasses, dsExceptionInfo);
  if( dynlen(dsClasses) > 0 )
  {
    iLen = dynlen(dsClasses);
    for( iClassesIndex = 1 ; iClassesIndex <= iLen ; iClassesIndex++ )
    {
      if( strpos(dsClasses[iClassesIndex], CPC_CONFIG_ALARM_ACK_POSTFIX) > 0 )
      {
        return "TRUE";
      }
    }
  }

  return "FALSE";

}//  cpcExportGenericFunctions_get5AlertAcknowledgeAlarmValue()





/**Add is 5-range alert acknowledgeable to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@todo unify format: rename alarmDP to dpe and move to the end of parameters

@param sDp current dp name
@param sAlarmDp name of acknowledgeable's dpe
@param dsDpParameters object's output
@param dsExceptionInfo exception info
*/

void cpcExportGenericFunctions_get5AlertAcknowledgeAlarm(string sDp, string sAlarmDp, dyn_string &dsDpParameters, dyn_string &dsExceptionInfo)
{
  string sIsAck;


  sIsAck = cpcExportGenericFunctions_get5AlertAcknowledgeAlarmValue(sDp, sAlarmDp, dsExceptionInfo);
  dynAppend(dsDpParameters, sIsAck);

}//  cpcExportGenericFunctions_get5AlertAcknowledgeAlarm()





/**Add is alarm active to the device's output

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp current dp name
@param dsDpParameters object's output
@param sDpe name of alarm dpe, ProcessInput.PosSt by default
*/

void cpcExportGenericFunctions_getAlarmActive(string sDp, dyn_string &dsDpParameters, string sDpe = ".ProcessInput.PosSt")
{
  bool bIsAlarmActive;


  dpGet(sDp + sDpe + ":_alert_hdl.._active", bIsAlarmActive);
  dynAppend(dsDpParameters, bIsAlarmActive);

}//  cpcExportGenericFunctions_getAlarmActive()





/**Convertion dyn_string into string; elements is glued via separator

Function returns an empty string if input array is empty or string with all elements of array merged with separator.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@todo move to generic place. reuse in other places

@param dsArray data to process
@param sSeparator item's separator
@return sResult string
*/

string dynJoin(dyn_string dsArray, string sSeparator)
{
  int iArrayIndex, iLen;
  string sResult;


  if( dynlen(dsArray) == 0 )
  {
    return "";
  }
  else
  {
    sResult = dsArray[1];
    iLen    = dynlen(dsArray);

    for( iArrayIndex = 2; iArrayIndex <= iLen ; iArrayIndex++ )
    {
      sResult = sResult + sSeparator + dsArray[iArrayIndex];
    }

    return sResult;
  }

}//  dynJoin()





/**updates CPC_CONFIG_CONVERSION parameter based on _msg_conv dpe config

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  UI

@param sDp datapoint name
@param sDpe datapoint element
@param dsExceptionInfo for errors
*/

void cpcExportGenericFunctions_updateConversionParameters(string sDp, string sDpe, dyn_string dsExceptionInfo)
{
  bool bConfigExists;
  int iConversionType, iOrder;
  string sValue;
  dyn_float dfArguments;


  fwConfigConversion_get(sDp + sDpe, bConfigExists, DPCONFIG_CONVERSION_RAW_TO_ENG_MAIN, iConversionType, iOrder, dfArguments, dsExceptionInfo);
  if( bConfigExists )
  {
    if( iOrder == 1 )
    {
      sValue = "linear|" + dfArguments[1] + "|" + dfArguments[2];
      cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_CONVERSION, sValue, dsExceptionInfo);
    }
  }

}//  cpcExportGenericFunctions_updateConversionParameters()





void cpcExportGenericFunctions_updateAlarmParameters(string sDp, string sDpe, string sDeviceTypeName, dyn_string dsExceptionInfo)
{
  bool bIsAlertActive, bConfigExists;
  int iAlertType, iAlertClassIndex, iLen;
  string sAlertPanel, sAlertHelp, sAlarmAck, sLimits, sAlertLimit, sAlarmMessage;
  dyn_string dsAlertTexts, dsAlertClass, dsSummaryDpeList, dsAlertPanelParameters, dsAlertLimits, dsDpParameters, dsAlarmConfigLow, dsAlarmConfigMedium, dsAlarmConfigHigh, dsAlarmConfigSafety, dsSMSConfig;
  dyn_dyn_string ddsCategories;


  sAlarmAck = "TRUE";


  // retrieve alert details
  fwAlertConfig_get(sDp + sDpe, bConfigExists, iAlertType, dsAlertTexts, dsAlertLimits, dsAlertClass, dsSummaryDpeList, sAlertPanel, dsAlertPanelParameters, sAlertHelp, bIsAlertActive, dsExceptionInfo);

  if( iAlertType == DPCONFIG_ALERT_NONBINARYSIGNAL )
  {
    if( sDeviceTypeName == UN_CONFIG_CPC_ANALOGSTATUS_DPT_NAME )
    {
      // for AnalogStatus
      cpcExportGenericFunctions_get5AlertLimits(sDp, dsDpParameters);

      sLimits = strjoin(dsDpParameters, "|");
      if( sLimits != "" )
      {
        sLimits = sLimits + "|";
      }

      if( dsDpParameters[1] != "" ||
          dsDpParameters[4] != ""   )
      {
        // alarm Ack should be updated ONLY if there are LL or HH alarms defined, otherwise - keep the original value
        sAlarmAck = cpcExportGenericFunctions_get5AlertAcknowledgeAlarmValue(sDp, dsExceptionInfo);
      }

      if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_VALUES, dsExceptionInfo) || sLimits != "" )
      {
        cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_VALUES, sLimits, dsExceptionInfo);
      }
    }
    else
    {
      if( sDeviceTypeName == UN_CONFIG_CPC_WORDSTATUS_DPT_NAME )
      {
        // for WordStatus
        iLen = dynlen(dsAlertClass);
        for( iAlertClassIndex = 1 ; iAlertClassIndex <= iLen ; iAlertClassIndex++ )
        {
          dpGet(sDp + sDpe + ":_alert_hdl." + iAlertClassIndex + "._match", sAlertLimit);   // retrieve limits separately, as _fwAlertConfig_get does not handle discrete alarms correctly
          if( strpos(dsAlertClass[iAlertClassIndex], "cpcP1") >= 0 )
          {
            dynAppend(dsAlarmConfigSafety, sAlertLimit);
          }
          else
          {
            if( strpos(dsAlertClass[iAlertClassIndex], "cpcP2") >= 0 )
            {
              dynAppend(dsAlarmConfigHigh, sAlertLimit);
            }
            else
            {
              if( strpos(dsAlertClass[iAlertClassIndex], "cpcP3") >= 0 )
              {
                dynAppend(dsAlarmConfigMedium, sAlertLimit);
              }
              else
              {
                if( strpos(dsAlertClass[iAlertClassIndex], "cpcP4") >= 0 )
                {
                  dynAppend(dsAlarmConfigLow, sAlertLimit);
                }
              }
            }
          }

          // alarm Ack should be updated ONLY if there are Safety (P1) or High (P2) alarms defined, otherwise - keep the original value
          if( ((strpos(dsAlertClass[iAlertClassIndex], "cpcP1") >= 0) || strpos(dsAlertClass[iAlertClassIndex], "cpcP2") >= 0 ) &&
              (strpos(dsAlertClass[iAlertClassIndex], "_Ack") < 0) )
          {
            sAlarmAck = "FALSE";
          }
        }


        if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_VALUES_LOW, dsExceptionInfo) || dynlen(dsAlarmConfigLow) > 0 )
        {
          cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_VALUES_LOW, strjoin(dsAlarmConfigLow, "|"), dsExceptionInfo);
        }

        if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_VALUES_MEDIUM, dsExceptionInfo) || dynlen(dsAlarmConfigMedium) > 0 )
        {
          cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_VALUES_MEDIUM, strjoin(dsAlarmConfigMedium, "|"), dsExceptionInfo);
        }

        if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_VALUES_HIGH, dsExceptionInfo) || dynlen(dsAlarmConfigHigh) > 0 )
        {
          cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_VALUES_HIGH, strjoin(dsAlarmConfigHigh, "|"), dsExceptionInfo);
        }

        if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_VALUES_SAFETY, dsExceptionInfo) || dynlen(dsAlarmConfigSafety) > 0 )
        {
          cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_VALUES_SAFETY, strjoin(dsAlarmConfigSafety, "|"), dsExceptionInfo);
        }
      }
    }


    unGenericDpFunctions_getKeyDeviceConfiguration(sDp, CPC_CONFIG_SMS_MESSAGE, dsSMSConfig, dsExceptionInfo);
    sAlarmMessage = dynlen(dsSMSConfig) > 0 ? dsSMSConfig[1] : "";

    if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_MESSAGE, dsExceptionInfo) || sAlarmMessage != "" )
    {
      cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_MESSAGE, sAlarmMessage, dsExceptionInfo);
    }

    if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_ACK, dsExceptionInfo) || sAlarmAck != "TRUE" )
    {
      cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_ACK, sAlarmAck, dsExceptionInfo);
    }

    if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_ACTIVE, dsExceptionInfo) || !bIsAlertActive )
    {
      cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_ACTIVE, bIsAlertActive?"true":"false", dsExceptionInfo, dsExceptionInfo);
    }


    unProcessAlarm_getCategory(makeDynString(sDp + sDpe), ddsCategories, dsExceptionInfo);
    if( cpcGenericDpFunctions_doesDevicePropertyExist(sDp, CPC_CONFIG_ALARM_SMS, dsExceptionInfo) || dynlen(ddsCategories[1]) > 0 )
    {
      cpcGenericDpFunctions_setDeviceProperty(sDp, CPC_CONFIG_ALARM_SMS, strjoin(ddsCategories[1],"|"), dsExceptionInfo);
    }
  }

}//  cpcExportGenericFunctions_updateAlarmParameters()
