/**@file test_unRcpDbFunctions.ctl  
   
@author: Juan Manuel Pablo Rodriguez (BE-ICS-FT)

@par Creation Date
/		 17/12/2021
 */
 
// You must pick up this library:
#uses "fwUnitTestComponentAsserts.ctl"

// and, of course, the library of routines you are wanting to call (ie wanting to test):
#uses "unRcpFunctions.ctl"
#uses "unRcpDbFunctions.ctl"

dyn_string exceptionInfo;
dyn_string dsRcpDbList = makeDynString();
string sDpInstanceNameTest = "test_unRecipeDbFunctions_saveInstance";
string className = "Test_Class_Name";
string instanceName = "TestInstanceName";
string comment = "Test comment description";
string sRcpName = className + UN_RECIPE_DELIMETER_CHAR + instanceName;
string sRcpDp = sDpInstanceNameTest + ".";

//-------------------------------------------------------
void testUnRecipe_unRecipeDbFunctions_setupSuite() {

}
//-------------------------------------------------------
void testUnRecipe_unRecipeDbFunctions_setup() {
  // Create a dummy UnRcpInstance datapoint
  if(!dpExists(sDpInstanceNameTest)) {
	int iRetVal = dpCreate(sDpInstanceNameTest, "UnRcpInstance");
	assertEqual(0, iRetVal);
  }
}
//-------------------------------------------------------
/** 
 * Try to save a recipe instance when a device has been locked previously.
 */
void testUnRecipe_unRecipeDbFunctions_savelockedInstance() {
  dyn_string dsLockedDeviceDp, dsLockedDeviceAlias;

  bool isLocked = _unRecipeFunctions_lockDevices(makeDynString(sRcpDp), dsLockedDeviceDp, dsLockedDeviceAlias, exceptionInfo, true);
  assertTrue(isLocked, "Error locking the mock instance devices.");

  bool isSaved = _unRecipeFunctions_saveRecipeInstanceToDb(sRcpDp, className, instanceName, comment, dsRcpDbList, exceptionInfo, true);
  assertTrue(!isSaved, "Saving the recipe should not be possible for an instance that is currently locked.");
  assertEqual("Error: The recipe instance 'Test_Class_Name/TestInstanceName' could not be locked. The recipe "
			+ "will not be saved to the ORACLE database.", exceptionInfo[2], "The exceptionInfo should have "
			+ "returned that the instance cannot be locked and the recipe will not be saved.");

  //unlock recipe instance for further tests
  bool isUnlocked = _unRecipeFunctions_unlockDevices(makeDynString(sRcpDp), exceptionInfo, true);
  assertTrue(isUnlocked, "Error unlocking the mock instance devices before finishing the test case.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
/** 
 * Try to save a recipe instance when a the state is different than blank or UN_RCP_SAVE_TO_DB_STATE.
 */
void testUnRecipe_unRecipeDbFunctions_saveBadStateInstance() {
  int rc;
  rc = dpSet(sRcpDp + "ProcessInput.State", UN_RCP_WARNING_MESSAGE_VALUES_MODIFIED);
  assertTrue(rc==0, "Error while setting up the state of the recipe instance to a mocked warning for values being modified.");

  bool isSaved = _unRecipeFunctions_saveRecipeInstanceToDb(sRcpDp, className, instanceName, comment, dsRcpDbList, exceptionInfo, true);
  assertTrue(!isSaved, "Saving the recipe should not be possible for an instance whose state is not blank or UN_RCP_SAVE_TO_DB_STATE");
  assertEqual("Error: The recipe instance 'Test_Class_Name/TestInstanceName' is in state Warning! The recipe values have been modified. "
			+ "Please reload the recipe to avoid inconsistencies.. The recipe will not be saved to the ORACLE database.", exceptionInfo[2], 
			"The exceptionInfo should have returned that the instance cannot be saved to the Database because is in state UN_RCP_WARNING_MESSAGE_VALUES_MODIFIED.");
  
  //reset the state of the recipe instance to blank
  rc=dpSet(sRcpDp + "ProcessInput.State", "");
  assertTrue(rc==0, "Error while resetting the state of the recipe instance to blank.");
  //unlock recipe instance for further tests
  bool isUnlocked = _unRecipeFunctions_unlockDevices(makeDynString(sRcpDp), exceptionInfo, true);
  assertTrue(isUnlocked, "Error unlocking the mock instance devices before finishing the test case.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
/** 
 * Try to save a recipe instance when recipe object does not exist.
 */
void testUnRecipe_unRecipeDbFunctions_saveInstanceCannotLoad() {
  bool isSaved = _unRecipeFunctions_saveRecipeInstanceToDb(sRcpDp, className, instanceName, comment, dsRcpDbList, exceptionInfo, true);
  assertTrue(!isSaved, "Saving the recipe should not be possible for an instance whose objects does not exist");
  assertEqual("Error getting the data of the recipe instance: 'Test_Class_Name/TestInstanceName'. The recipe instance cannot be saved to the ORACLE database.", 
			exceptionInfo[2], "The exceptionInfo should have returned that the recipe instance cannot be saved in the Database because of " +
			"an error getting the data from the object.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
/** 
 * Check save instance error when devices cannot be locked.
 */
void testUnRecipe_unRecipeDbFunctions_checkSaveInstanceErrorCodeCannotLock() {
  _unRecipeFunctions_finalizeSaveRecipeToDb(DB_SAVE_ERROR_CANNOT_LOCK_INSTANCE, sRcpName, sRcpDp, "", exceptionInfo, true);
  assertEqual("Error: The recipe instance 'Test_Class_Name/TestInstanceName' could not be locked. The recipe will not be saved to the ORACLE database.", 
			exceptionInfo[2], "The exceptionInfo should have returned that the recipe instance cannot be locked in the Database.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
/** 
 * Check save instance error when instance has incorrect state.
 */
void testUnRecipe_unRecipeDbFunctions_checkSaveInstanceErrorCodeBadState() {
  _unRecipeFunctions_finalizeSaveRecipeToDb(DB_SAVE_ERROR_BAD_RCP_INSTANCE_STATE, sRcpName, sRcpDp, UN_RCP_WARNING_MESSAGE_VALUES_MODIFIED, exceptionInfo, true);
  assertEqual("Error: The recipe instance 'Test_Class_Name/TestInstanceName' is in state Warning! The recipe values have been modified. Please reload the recipe to avoid inconsistencies.. "
			+ "The recipe will not be saved to the ORACLE database.", exceptionInfo[2], "The exceptionInfo should have returned that "
			+ "the recipe instance cannot be saved to the Database because of an incorrect instance state.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
/** 
 * Check save instance error when instance cannot be loaded.
 */
void testUnRecipe_unRecipeDbFunctions_checkSaveInstanceErrorCodeCannotLoad() {
  _unRecipeFunctions_finalizeSaveRecipeToDb(DB_SAVE_ERROR_CANNOT_LOAD_INSTANCE, sRcpName, sRcpDp, UN_RCP_SAVE_TO_DB_STATE, exceptionInfo, true);
  assertEqual("Error getting the data of the recipe instance: 'Test_Class_Name/TestInstanceName'. The recipe instance cannot be saved to the ORACLE database.",
			  exceptionInfo[2], "The exceptionInfo should have returned that the recipe instance cannot be saved to the Database because of an "
			  + "error retrieving the data of the instance.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
/** 
 * Check save instance error when instance cannot be reset.
 */
void testUnRecipe_unRecipeDbFunctions_checkSaveInstanceErrorCodeCannotReset() {
  _unRecipeFunctions_finalizeSaveRecipeToDb(DB_SAVE_ERROR_CANNOT_RESET_INSTANCE, sRcpName, sRcpDp, UN_RCP_SAVE_TO_DB_STATE, exceptionInfo, true);
  assertEqual("Error while resetting the recipe instance 'Test_Class_Name/TestInstanceName' in the ORACLE database.",
			  exceptionInfo[2], "The exceptionInfo should have returned that the recipe instance cannot be saved to the Database because of an "
			  + "error resetting the recipe instance.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
/** 
 * Check save instance error when instance cannot be saved.
 */
void testUnRecipe_unRecipeDbFunctions_checkSaveInstanceErrorCodeCannotSave() {
  _unRecipeFunctions_finalizeSaveRecipeToDb(DB_SAVE_ERROR_CANNOT_SAVE_INSTANCE, sRcpName, sRcpDp, UN_RCP_SAVE_TO_DB_STATE, exceptionInfo, true);
  assertEqual("Error saving the recipe instance 'Test_Class_Name/TestInstanceName' to the ORACLE database.", exceptionInfo[2], 
			 "The exceptionInfo should have returned that the recipe instance cannot be saved to the Database because of an error.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
/** 
 * Check save instance message when instance is saved successfully.
 */
void testUnRecipe_unRecipeDbFunctions_checkSaveInstanceMessageSuccessCode() {
  _unRecipeFunctions_finalizeSaveRecipeToDb(DB_SAVE_RCP_INSTANCE_SAVED, sRcpName, sRcpDp, UN_RCP_SAVE_TO_DB_STATE, exceptionInfo, true);
  assertEqual("The recipe instance 'Test_Class_Name/TestInstanceName' was saved successfully to the ORACLE database.", exceptionInfo[2], 
			 "The exceptionInfo should have returned that the recipe instance has been saved to the Database.");
  dynClear(exceptionInfo);
}

//-------------------------------------------------------
void testUnRecipe_unRecipeDbFunctions_teardown() {

}
//-------------------------------------------------------
void testUnRecipe_unRecipeDbFunctions_teardownSuite() {
  // Delete the dummy datapoint
  if (dpExists(sDpInstanceNameTest)) {
    int iRetVal = dpDelete(sDpInstanceNameTest);
    assertEqual(0, iRetVal);
  }
}
